#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module for data exporting into .csv or hdf5 format. The module is called from runner.py, reads-in a
configuration file and exports data into one of both mentioned formats.
"""

import inspect
import os
from datetime import datetime
from typing import Any

import pandas as pd
from dateutil import parser as date_parser
from sqlalchemy.orm.session import Session

from model.database import tables
from model.database.db_handler import DatabaseHandler
from model.database.tables import metadata
from model.utilities.time_helper import TimeHelper
from model.utilities.utilities import read_config, load_program_config


def database_session(filename: str = None, db_path: str = None) -> Session:
    """ Returns an open SqlAlchemy-Session. The session is retrieved from the DatabaseHandler.

    @param filename: Name of the configuration file to init the DatabaseHandler
    @param db_path: Path to the database. Default: current working directory
    @return: SqlAlchemy-Session
    """
    min_return_tuples = load_program_config().get("request_settings").get("min_return_tuples", 1)
    db_handler = DatabaseHandler(metadata=metadata, path=db_path, min_return_tuples=min_return_tuples,
                                 **read_config(file=filename, section="database"))
    return db_handler.session_factory()


class CsvExport:
    """ Class to actually query and save data. The file-format is given as input parameter, along with *args and
    **kwargs for the pd.to_csv(*args, **kwargs) and pd.to_hdf(*args, **kwargs).
    """

    def __init__(self, file: str = None):
        self.config = read_config(file=file, section=None)
        self.db_handler = DatabaseHandler(metadata, **self.config.get("database"))
        self.options = self.config["query_options"]
        self.filename = f"{self.options.get('table_name')}_{TimeHelper.now().strftime('%Y-%m-%dT%H-%M-%S')}"
        self.path = os.getcwd()

        # extract and convert starting point
        self.from_timestamp: str = self.options.get("from_timestamp", None)
        if self.from_timestamp:
            self.from_timestamp = date_parser.parse(self.from_timestamp, dayfirst=True)
        else:
            self.from_timestamp = datetime.min

        # extract and convert ending point
        self.to_timestamp: str = self.options.get("to_timestamp", None)
        if self.to_timestamp and self.to_timestamp != "now":
            self.to_timestamp = date_parser.parse(self.to_timestamp, dayfirst=True)
        else:
            self.to_timestamp = TimeHelper.now()

        # get table object to pass over as an argument in self.create_csv()
        table_names = dict()
        for name, obj in inspect.getmembers(tables):
            if inspect.isclass(obj):
                table_names.update({name: obj})
        self.table = table_names[self.options.get("table_name", None)]

    def load_data(self) -> pd.DataFrame:
        """ Receives from the DatabaseHandler the tuples that should be exported.
        The received tuples are based on the parameters set by the user in csv-config.yaml.
        The method tries to create the save-path if it not already exists.
        Creates or modifies the file.
        All previously stored content in the file will be erased.
        """
        ticker_data = self.db_handler.get_readable_query(self.table,
                                                         self.options.get("query_everything", None),
                                                         self.from_timestamp,
                                                         self.to_timestamp,
                                                         self.options.get("exchanges", []),
                                                         self.options.get("currency_pairs", []),
                                                         self.options.get("first_currencies", []),
                                                         self.options.get("second_currencies", [])
                                                         )
        return pd.DataFrame(ticker_data)

    def export(self, data_type: str = "csv", *args: Any, **kwargs: Any) -> Any:
        """ Exports the data in the specified format.

        @param data_type: String representation of the export format. Default: csv.
        """

        ticker_data = self.load_data()
        if self.filename.endswith(".csv"):
            output_path: str = os.path.join(self.path, self.filename)
        else:
            output_path: str = os.path.join(self.path, f"{self.filename}.csv")

        export_format = {"csv": {"function": ticker_data.to_csv,
                                 "parameters": ["path_or_buf", "sep", "decimal", "index"]},
                         "hdf": {"function": ticker_data.to_hdf,
                                 "parameters": ["path_or_buf"]}}

        parameters = {"path_or_buf": output_path,
                      "sep": self.options.get("delimiter", ";"),
                      "decimal": self.options.get("decimal", "."),
                      "index": False,
                      }

        parameters = {k: v for k, v in parameters.items() if k in export_format.get(data_type).get("parameters")}
        parameters.update(**kwargs)

        export_format.get(data_type, "csv").get("function")(*args, **parameters)
        print(output_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Validators to validate the request template.

The module checks for existence, types and values of necessary parameters.

Validators structure:

- RequestTemplateValidator
  - LoadFileValidator
  - LoadYamlValidator
  - RequestTemplateCoreValidator
    - ConfigSectionValidator
    - DatabaseStringValidator
    - OperationSettingKeyValidator
    - RequestKeysValidator
"""
import sys
from typing import Any, Text, Dict

from model.validating.api_map_validators import LoadFileValidator, LoadYamlValidator
from model.validating.base import Report, Validator, CompositeValidator
from model.validating.errors import KeyNotInDictError, WrongTypeError, WrongValueError


class RequestTemplateValidator(CompositeValidator):
    """
    Validator for an request template.

    Attributes:
        - value: The file name or file path as Text.
    """
    value: Text

    def validate(self) -> bool:
        """
        Validates the value.

        @return: Whether further Validators may continue validating.
        """
        load_file = LoadFileValidator(self.value)
        is_file_loaded = load_file.validate()
        self.append_report(load_file)

        if not is_file_loaded:
            return False

        load_yaml = LoadYamlValidator(load_file)
        is_yaml_loaded = load_yaml.validate()
        self.append_report(load_yaml)

        if not is_yaml_loaded:
            return False

        request_template = RequestTemplateCoreValidator(load_yaml.get_result_value())
        can_continue = request_template.validate()
        try:
            for report in request_template.report.reports:
                self.append_report(report)
        except (TypeError, Exception):
            self.append_report(request_template)

        return can_continue

    def result(self) -> bool:
        """
        TODO: Fill out
        """
        if not self.report:
            self.validate()

        return True


class RequestTemplateCoreValidator(CompositeValidator):
    """
    Validator for the contents of an request template.
    """

    def __init__(self, value: Dict[Text, Any]):
        """
        Constructor of RequestTemplateCoreValidator.

        @param value: The value to get validated.
        """
        super().__init__(
            value,
            ConfigSectionValidator(value),
            DatabaseStringValidator(value["general"]["database"]),
            OperationSettingKeyValidator(value["general"]["operation_settings"]),
            RequestKeysValidator(value["jobs"])
        )


class ConfigSectionValidator(Validator):
    """
    Validator to check if all sections and blocks are present.
    """
    block = ["general", "jobs"]
    section = ["database", "operation_settings", "utilities"]

    def validate(self) -> bool:
        try:
            for key in self.value:
                if key not in ConfigSectionValidator.block:
                    raise KeyNotInDictError(key, dict.fromkeys(ConfigSectionValidator.block))

            for key in self.value.get("general"):
                if key not in ConfigSectionValidator.section:
                    raise KeyNotInDictError(key, dict.fromkeys(ConfigSectionValidator.section))
        except KeyNotInDictError as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report(f"Configuration contains all blocks: {ConfigSectionValidator.block} and"
                                 f" sections: {ConfigSectionValidator.section}")
            return True


class DatabaseStringValidator(Validator):
    """
    Validator to check if the database block is valid.
    """
    db_strings = (
        ("sqltype", str, "sqlite"),
        ("client", (str, type(None)), None),
        ("user_name", (str, type(None)), None),
        ("password", (str, type(None)), None),
        ("host", str, "localhost"),
        ("port", int, 5432),
        ("db_name", str, "ExampleDB")
    )

    def validate(self) -> bool:
        """
        Validates the value.

        @return: Whether further Validators may continue validating.
        """
        try:
            for (key, ex_type, ex_value) in DatabaseStringValidator.db_strings:
                if key not in self.value:
                    raise KeyNotInDictError(key, self.value)

                if not isinstance(self.value[key], ex_type):
                    raise WrongTypeError(ex_type, type(self.value[key]), key)

                if self.value[key] != ex_value:
                    raise WrongValueError(ex_value, self.value[key], key)
        except (KeyNotInDictError, WrongTypeError) as error:
            self.report = Report(error)
            return False
        else:
            self.report = Report("Database connection string is valid")
            return True


class OperationSettingKeyValidator(Validator):
    """
    Validator to check if the operation settings block is valid.
    """
    db_strings = (
        ("frequency", str, "once"),
        ("interval", str, "days"),
        ("timeout", int, 10),
        ("enable_logging", bool, True),
        ("asynchronously", bool, True)
    )

    def validate(self) -> bool:
        """
        Validates the value.

        @return: Whether further Validators may continue validating.
        """
        try:
            for (key, ex_type, ex_value) in OperationSettingKeyValidator.db_strings:
                if key not in self.value:
                    raise KeyNotInDictError(key, self.value)

                if not isinstance(self.value[key], ex_type):
                    raise WrongTypeError(ex_type, type(self.value[key]), key)

                if self.value[key] != ex_value:
                    raise WrongValueError(ex_value, self.value[key], key)
        except (KeyNotInDictError, WrongTypeError) as error:
            self.report = Report(error)
            return False
        else:
            self.report = Report("Operation settings are valid")
            return True


class RequestKeysValidator(Validator):
    """
    Validates if all keys exist and types are correct for the request itself.
    """
    sections = (
        ("request_method", type(None), None),
        ("update_cp", bool, False),
        ("excluded", type(None), None),
        ("exchanges", type(None), None),
        ("currency_pairs", type(None), None),
        ("first_currencies", type(None), None),
        ("second_currencies", type(None), None),
    )

    def validate(self) -> bool:
        """
        Validates the value.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "JobName" not in self.value:
                raise KeyNotInDictError("JobName", self.value)

            job = self.value["JobName"]
            for (key, ex_type, ex_value) in RequestKeysValidator.sections:
                if key not in job:
                    raise KeyNotInDictError(key, job)

                if not isinstance(job[key], ex_type):
                    raise WrongTypeError(ex_type, type(job[key]), key)

                if job[key] != ex_value:
                    raise WrongValueError(ex_value, job[key], key)
        except (KeyNotInDictError, WrongTypeError, WrongValueError) as error:
            self.report = Report(error)
            return False
        else:
            self.report = Report("Request keys and types are valid.")
            return True


if __name__ == "__main__":
    IS_VALID = RequestTemplateValidator("open_crypto/resources/templates/request_template.yaml").validate()
    sys.exit(int(not IS_VALID))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Contains custom database types.

Classes:
 - UnixTimestampMs: Database type to store dates and times as Unix timestamps in milliseconds (UTC+0).
"""

from datetime import datetime
from typing import Type, Optional

from sqlalchemy.engine import Dialect
from sqlalchemy.types import BIGINT, TypeDecorator

from model.utilities.time_helper import TimeHelper, TimeUnit


class UnixTimestampMs(TypeDecorator):
    """
    Database type to store Unix timestamps in milliseconds (UTC+0).

    Although it is stored as an integer internally, it is used like a timezone aware (UTC+0) datetime.
    """
    impl = BIGINT

    cache_ok = True

    @property
    def python_type(self) -> Type[datetime]:
        return datetime

    def process_bind_param(self, value: datetime, dialect: Dialect) -> Optional[int]:
        if value is None:
            return None

        return int(TimeHelper.to_timestamp(value, TimeUnit.MILLISECONDS))

    def process_result_value(self, value: int, dialect: Dialect) -> Optional[datetime]:
        if value is None:
            return None

        return TimeHelper.from_timestamp(value, TimeUnit.MILLISECONDS)

    def process_literal_param(self, value: int, dialect: Dialect) -> str:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Contains helper classes to manage dates and times.

Classes:
- TimeHelper: Used to create/convert timezone aware (UTC+0) dates and times.

Enums:
- TimeUnit: Used to indicate the unit of timestamps.
"""

from datetime import datetime, timezone
from enum import IntEnum

from datetime_periods import period
from dateutil.parser import parse


class TimeUnit(IntEnum):
    """
    An enumeration to indicate the unit of timestamps.
    """
    SECONDS = 0
    MILLISECONDS = 1
    MICROSECONDS = 2
    NANOSECONDS = 3


class TimeHelper:
    """
    A helper class to create/convert dates and times.

    It ensures that all dates and times are timezone aware (UTC+0).

    freq_map is used to convert specific strings from plural into singular.
    """
    freq_map = {
        "minutes": "minute",
        "hours": "hour",
        "days": "day",
        "weeks": "week",
        "months": "month"
    }

    @staticmethod
    def now() -> datetime:
        """
        Get the current datetime (UTC+0).

        The accuracy is limited to milliseconds and the remaining microseconds are cleared.

        @return: The current datetime (UTC+0).
        @rtype: datetime
        """
        now = datetime.now(tz=timezone.utc)
        return now.replace(microsecond=now.microsecond - now.microsecond % 1000)

    @staticmethod
    def now_timestamp(unit: TimeUnit = TimeUnit.SECONDS) -> float:
        """
        Get the timestamp of the current datetime (UTC+0).

        @param unit: The desired time unit of the timestamp.
        @type unit: TimeUnit

        @return: The timestamp of the current datetime (UTC+0).
        @rtype: float
        """
        return TimeHelper.to_timestamp(TimeHelper.now(), unit)

    @staticmethod
    def from_string(representation: str) -> datetime:
        """
        Get a datetime (UTC+0) from a given representation.

        @param representation: The string that represents a datetime.
        @type representation: str

        @return: The datetime (UTC+0) of the given representation.
        @rtype: datetime
        """
        return parse(representation).replace(tzinfo=timezone.utc)

    @staticmethod
    def from_timestamp(timestamp: float, unit: TimeUnit = TimeUnit.SECONDS) -> datetime:
        """
        Get a datetime (UTC+0) from a given timestamp.

        @param timestamp: The timestamp whose datetime is to be obtained.
        @type timestamp: float
        @param unit: The time unit in which the timestamp is given.
        @type unit: TimeUnit

        @return: The datetime (UTC+0) of the given timestamp.
        @rtype: datetime
        """
        timestamp_in_sec: float = timestamp / (1000 ** int(unit))
        return datetime.fromtimestamp(timestamp_in_sec, tz=timezone.utc)

    @staticmethod
    def to_timestamp(date_time: datetime, unit: TimeUnit = TimeUnit.SECONDS) -> float:
        """
        Convert a datetime to a timestamp.

        @param date_time: The datetime to be converted.
        @type date_time: datetime
        @param unit: The desired time unit of the timestamp.
        @type unit: TimeUnit

        @return: The timestamp of the given datetime in the desired time unit.
        @rtype: float
        """
        return date_time.replace(tzinfo=timezone.utc).timestamp() * (1000 ** int(unit))

    @staticmethod
    def start_end_conversion(date_time: datetime, frequency: str, to_end: bool = True) -> datetime:
        """
        Returns the beginning/end of a period.

        @param date_time: The datetime object to be converted.
        @type date_time: datetime
        @param frequency: The underlying period frequency.
        @type frequency: str
        @param to_end: boolean, return end of period. Default: True
        @type to_end: bool

        @return: datetime of start/end of period.
        @rtype: datetime
        """
        return period(date_time, TimeHelper.freq_map[frequency])[int(to_end)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module to find, extract and convert values from the exchange responses.

Classes:
    - Mapping
Functions:
    - convert_type
    - extract_mappings
    - is_scalar
"""

import logging
from collections import deque
from collections.abc import Iterable
from typing import Any, List, Dict, Deque, Tuple

from model.utilities.utilities import TYPE_CONVERSIONS


def convert_type(value: Any, types_queue: Deque[str]) -> Any:
    """
    Converts the value via type conversions.

    Helper method to convert the given value via a queue of type conversions.

    @param value: The value to get converted to another type.
    @type value: Any
    @param types_queue: The queue of type conversion instructions.
    @type types_queue: deque

    @return: The converted value.
    @rtype: Any
    """
    current_type = types_queue.popleft()

    result = value

    while types_queue:
        next_type = types_queue.popleft()

        types_tuple = (current_type, next_type)

        if "continue" in types_tuple:
            continue

        conversion = TYPE_CONVERSIONS[types_tuple]

        params = list()

        for _ in range(conversion["params"]):
            params.append(types_queue.popleft())

        # Change here to avoid "None" as result value in the params when no value to convert is needed (i.e. when
        # methods are called with ("none", ...).
        # if not result and isinstance(result, (str, list)):
        try:
            if result is None:
                result = conversion["function"](*params)
            else:
                result = conversion["function"](result, *params)

            current_type = next_type
        except Exception:
            return None

    return result


class Mapping:
    """
    Class representing mapping data and logic.

    Class representing mapping data und further functionality provided
    with methods.

    Attributes:
        key:
            String being the keyword indicating one or several
            database table columns. See "database_column_mapping"
            in "config.yaml".
        path:
            An ordered list of keys used for traversal through the
            response dict with the intention of returning the value wanted
            for the database.
        types:
            An ordered sequence of types and
            additional parameters (if necessary). Is used to conduct
            type conversions within the method "extract_value()".
    """

    def __init__(self,
                 key: str,
                 path: List[str],
                 types: List[str]):
        """
        Constructor of Mapping.

        Constructor method for constructing method objects.

        @param key: String being the keyword indicating one or several
                    database table columns. See "database_column_mapping"
                    in "config.yaml".
        @type key: str
        @param path: An ordered list of keys used for traversal through the
                     response dict with the intention of returning the value wanted
                     for the database.
        @type path: list
        @param types: An ordered sequence of types and
                      additional parameters (if necessary). Is used to conduct
                      type conversions within the method "extract_value()".
        @type types: list
        """
        self.key = key
        self.path = path
        self.types = types

    def traverse_path(self, response: Dict[str, Any], path_queue: Deque[str],
                      currency_pair_info: Tuple[str, str, str] = None) \
            -> Any:
        """
        Traverses the path on a response.

        Helper method for traversing the path on the given response dict (subset).

        @param response: The response dict (subset).
        @type response: dict
        @param path_queue: The queue of path traversal instructions.
        @type path_queue: deque
        @param currency_pair_info: The formatted String of a currency pair.
                                   For special case that the key of a dictionary is the formatted currency pair string.
        @type currency_pair_info: tuple[str, str, str]

        @return: The traversed response dict.
        @rtype: Optional[dict]
        """
        path_element = path_queue.popleft()

        if path_element == "dict_key":
            # Special case to extract value from "dict_key"
            traversed = list(response.keys())
        elif path_element == "dict_values":
            # Special case to extract value from "dict_values"
            traversed = list(response.values())
        elif path_element == "list_key":
            # Special case with the currency_pair prior to a list
            traversed = list(response.keys())
        elif path_element == "list_values":
            traversed = list(response.values())
        elif path_element == []:
            # Special case to extract multiple values from a single list ["USD","BTC",...]
            traversed = response
        elif path_element == "currency_pair" and currency_pair_info[2] is not None:
            traversed = response[currency_pair_info[2]]
        elif is_scalar(response):
            return None
        else:  # Hier editiert für Kraken sonderfall
            if isinstance(response, dict) and path_element not in response.keys():
                return None
            else:
                traversed = response[path_element]

        return traversed

    def extract_value(self,
                      response: List[Dict[str, Any]],
                      path_queue: Deque[str] = None,
                      types_queue: Deque[str] = None,
                      iterate: bool = True,
                      currency_pair_info: Tuple[str, str, str] = (None, None, None)) -> Any:
        """
        Extracts the value specified by "self.path".

        Extracts the value specified by the path sequence and converts it
        using the "types" specified.

        @param response: The response dict (JSON) returned by an API request.
        @type response: Collection
        @param path_queue: The queue of path traversal instructions.
        @type path_queue: deque
        @param types_queue: The queue of type conversion instructions.
        @type types_queue: deque
        @param iterate: Whether still an auto-iteration is possible.
        @type iterate: bool
        @param currency_pair_info: The formatted String of a currency pair.
        @type currency_pair_info: tuple[str, str, str]

        @return: The value specified by "path_queue" and converted
                 using "types_queue".
                 Can be a list of values which get extracted iteratively from
                 the response.
        @rtype: Any
        """
        if path_queue is None:
            path_queue = deque(self.path)

        if types_queue is None:
            types_queue = deque(self.types)

        if not response:
            return None

        if not path_queue:
            # TODO: after integration tests, look if clause for first and second currency can be deleted!
            if types_queue[0] == "first_currency":
                return currency_pair_info[0]
            elif types_queue[0] == "second_currency":
                return currency_pair_info[1]
            return convert_type(None, types_queue)

        while path_queue:

            if iterate and isinstance(response, list):
                # Iterate through list of results
                result = list()

                # special case for bitfinex
                if len(response) == 1:
                    response = response[0]
                    continue  # because instance of response has to be checked

                for item in response:

                    if is_scalar(item):
                        return self.extract_value(response,
                                                  path_queue,
                                                  types_queue,
                                                  iterate=False)

                    result.append(
                        self.extract_value(
                            item,
                            deque(path_queue),
                            deque(types_queue)
                        )
                    )

                return result

            elif is_scalar(response):
                # Return converted scalar value
                return convert_type(response, types_queue)

            # Special case for Bitz to handle empty dicts/lists.
            elif not response:
                return None

            else:
                # Traverse path
                response = self.traverse_path(response, path_queue, currency_pair_info=currency_pair_info)

        if types_queue and response is not None:  # None to allow to change 0 to boolean.

            if isinstance(response, list):

                result = list()

                for item in response:
                    result.append(
                        convert_type(item, deque(types_queue))
                    )

                # for dict_key special_case aka. test_extract_value_list_containing_dict_where_key_is_value() in test_mapping.py
                if len(result) == 1:
                    result = result[0]

                response = result

            else:
                response = convert_type(response, types_queue)

        return response

    def __str__(self) -> str:
        """String representation of a Mapping"""
        string_path = list()

        for item in self.path:
            string_path.append(str(item))

        return " / ".join(string_path) + " -> " + str(self.key)


def extract_mappings(exchange_name: str, requests: Dict[str, Any]) -> Dict[str, List[Mapping]]:
    """
    Helper-Method which should be only called by the constructor.
    Extracts out of a given exchange .yaml-requests-section for each
    request the necessary mappings so the values can be extracted from
    the response for said request.

    The key-value in the dictionary is the same as the key for the request.
    i.e. behind 'ticker' are all the mappings stored which are necessary for
    extracting the values out of a ticker-response.

    If there is no mapping specified in the .yaml for a value which is contained
    by the response, the value will not be extracted later on because there won't
    be a Mapping-object for said value.

    @param exchange_name: str
        String representation of the exchange name.
    @param requests: Dict[str: List[Mapping]]
        Requests-section from a exchange.yaml as dictionary.
        Method does not check if dictionary contains viable information.

    @return:
        Dictionary with the following structure:
            {'request_name': List[Mapping]}
    """
    response_mappings = dict()
    if requests:
        for request in requests:
            request_mapping = requests[request]

            if "mapping" in request_mapping.keys():
                mapping = request_mapping["mapping"]
                mapping_list = list()

                try:
                    for entry in mapping:
                        mapping_list.append(Mapping(entry["key"], entry["path"], entry["type"]))
                except KeyError:
                    print(f"Error loading mappings of {exchange_name} in {request}: {entry}")
                    logging.error("Error loading mappings of %s in %s: %s", exchange_name, request, entry)
                    break

                response_mappings[request] = mapping_list

    return response_mappings


def is_scalar(value: Any) -> bool:
    """
    Indicates whether a value is a scalar or not.

    Convenience function returning a bool whether the provided value is a single value or not.
    Strings count as scalar although they are iterable.

    @param value: The value to evaluate concerning whether it is a single value
                  or multiple values (iterable).
    @type value: Any

    @return: Bool indicating whether the provided value is a single value or not.
    @rtype: bool
    """
    return isinstance(value, str) or not isinstance(value, Iterable)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Exchange API Map Validation

Classes for validating the API yaml files of Exchanges and generating
a complete Report.

general structure:
( validate.py calls ApiMapValidator )
-ApiMapValidator line 689 (calls)
    -LoadFileValidator
    -LoadYamlValidator
    -ApiMapValidator (contains)
        -NameValidator
        -ApiUrlValidator (calls)
            -UrlValidator
        -RateLimitValidator
        -RequestsValidator (calls)
            -ApiMethodValidator (contains)
                -RequestValidator (contains)
                    -TemplateValidator
                    -PairTemplateValidator
                    -ParamsValidator (calls)
                        -ParamValidator
                -ResponseValidator
                -MappingValidator (calls)
                    -MappingEntryValidator (contains)
                        -KeyValidator
                        -PathValidator
                        -TypeValidator
        -RequestMappingValidator

Authors:
    Carolina Keil
    Martin Schorfmann
    Paul Böttger

Since:
    19.03.2019

Version:
    26.04.2019
"""
import re
from typing import Any, Text, Dict

import oyaml as yaml
import validators

from model.database.tables import ExchangeCurrencyPair, Ticker, HistoricRate, OrderBook, Trade
from model.validating.base import Report, CompositeReport, Validator, CompositeValidator, ProcessingValidator
from model.validating.errors import KeyNotInDictError, SubstringNotInStringError, WrongTypeError, UrlValidationError, \
    NamingConventionError


# pylint: disable=too-many-lines


class ApiMapFileValidator(CompositeValidator):
    """
    Validator for validating a given API Map file.

    Consists of a FileLoadValidator, a YamlLoadValidator and an ApiMapValidator.

    Attributes:
        value:
            The file name or file path as Text.
    """

    value: Text

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        load_file = LoadFileValidator(self.value)
        is_file_loaded = load_file.validate()
        self.append_report(load_file)

        if not is_file_loaded:
            return False

        load_yaml = LoadYamlValidator(load_file)
        is_yaml_loaded = load_yaml.validate()
        self.append_report(load_yaml)

        if not is_yaml_loaded:
            return False

        api_map = ApiMapValidator(load_yaml.get_result_value())
        can_continue = api_map.validate()
        self.append_report(api_map)

        return can_continue


class LoadFileValidator(ProcessingValidator):
    """
    Validator for opening a file and reading it as Text.
    """

    def process(self) -> Any:
        """
        Processes the value.

        @return: The result value from processing the initial value.
        """
        with open(self.value, "r", encoding="UTF-8") as file:
            return file.read()

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            super().validate()
        except IOError as io_error:
            self.report = Report(io_error)
            return False
        else:
            self.report = Report("Load file was valid.")
            return True


class LoadYamlValidator(ProcessingValidator):
    """
    Validator for loading a YAML String.
    """

    def process(self) -> dict:
        """
        Processes the value.

        @return: The result value from processing the initial value.
        """
        return yaml.safe_load(self.value)

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            super().validate()
        except yaml.YAMLError as yaml_error:
            self.report = Report(yaml_error)
            return False
        else:
            self.report = Report("YAML Parsing successful.")
            return True


class ApiMapValidator(CompositeValidator):
    """
    Validator for calling the other required Validators.
    """

    def __init__(self, value: Dict[Text, Any]):
        """
        Constructor of ApiMapValidator.

        @param value: The value to get validated.
        """
        super().__init__(
            value,
            NameValidator(value),
            ApiUrlValidator(value),
            RateLimitValidator(value),
            RequestsValidator(value),
            RequestMappingValidator(value["requests"])
        )


class NameValidator(Validator):
    """
    Validator to validate a given exchange with respect to the 'name' key.

    Checks if 'name' exists in the file, is a string and conforms to the naming convention.

    Attributes:
        name_regex:
            The Regex pattern for valid names must meet.
            i.e. 'exchange', 'crypto_exchange'
    """
    name_pattern: Text = r"[a-z0-9\-_]+(\.){0,1}[a-z0-9]+"

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "name" not in self.value:
                raise KeyNotInDictError("name", self.value)
        except KeyNotInDictError as error:
            has_name = Report(error)
            self.report = has_name
        else:
            has_name = Report("Key 'name' exists.")
            name = self.value.get("name")
            try:
                if not isinstance(name, str):
                    raise WrongTypeError(str, type(name))
            except WrongTypeError as error:
                is_name_string = Report(error)
                self.report = CompositeReport(has_name, is_name_string)
            else:
                is_name_string = Report("Value of the 'name' key is a string.")
                try:
                    if not re.fullmatch(self.name_pattern, name):
                        raise NamingConventionError(self.name_pattern, name)
                except NamingConventionError as error:
                    meets_convention = Report(error)
                else:
                    meets_convention = Report("Naming convention met.")

                self.report = CompositeReport(has_name, is_name_string, meets_convention)

        return True


class UrlValidator(Validator):
    """
    Validator for validating a given URL.

    Checks whether the URL is a String and is valid.
    """

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if not isinstance(self.value, str):
                raise WrongTypeError(str, type(self.value))
        except WrongTypeError as error:
            is_url_str = Report(error)
            self.report = is_url_str
        else:
            is_url_str = Report("URL is a str.")
            try:
                url_validation_result = validators.url(self.value)
                if self.value != "" and not url_validation_result:
                    raise UrlValidationError(self.value, url_validation_result)
            except UrlValidationError as error:
                is_url_valid = Report(error)
            else:
                is_url_valid = Report("URL is valid.")

            self.report = CompositeReport(is_url_str, is_url_valid)

        return True


class ApiUrlValidator(Validator):
    """
    Validator for validating a given YAML file regarding the key 'api_url'.

    Checks whether 'api_url' exists in the root dict and is a valid URL
    according to an instance of the class UrlValidator.

    Attributes:
        value:
            The root dict that should contain the 'api_url' field.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "api_url" not in self.value:
                raise KeyNotInDictError("api_url", self.value)
        except KeyNotInDictError as error:
            has_api_url = Report(error)
            self.report = has_api_url
        else:
            has_api_url = Report("Key 'api_url' exists.")
            api_url = self.value.get("api_url")

            url_validator = UrlValidator(api_url)
            url_validator.validate()

            self.report = CompositeReport(has_api_url, url_validator.report)

        return True


class RateLimitValidator(Validator):
    """
    Validator for the field 'rate_limit'.

    Validator for validating the root or a part of the API Map
    regarding the field 'rate_limit'.
    Checks whether the optional key 'rate_limit' exists in the dict,
    the value of the key is None or is a dict.
    The fields of the dict are also checked.
    """

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        if "rate_limit" in self.value:
            is_rate_limit_key = Report("Optional key 'rate_limit' exists.")
            rate_limit = self.value.get("rate_limit")

            if rate_limit is None:
                has_rate_limit_value = Report("Value of optional key 'rate_limit' was None.")
                self.report = CompositeReport(
                    is_rate_limit_key,
                    has_rate_limit_value
                )
                return True
            has_rate_limit_value = Report("Value of optional key 'rate_limit' was not None.")

            try:
                if not isinstance(rate_limit, dict):
                    raise WrongTypeError(dict, type(rate_limit))
            except WrongTypeError as error:
                is_rate_limit_dict = Report(error)
                self.report = CompositeReport(
                    is_rate_limit_key,
                    has_rate_limit_value,
                    is_rate_limit_dict
                )
            else:
                is_rate_limit_dict = Report("Rate limit was a dict.")
                are_keys_valid = CompositeReport()

                for key in ["max", "unit"]:
                    try:
                        if key not in rate_limit:
                            raise KeyNotInDictError(key, rate_limit)
                    except KeyNotInDictError as error:
                        are_keys_valid.append_report(Report(error))
                    else:
                        are_keys_valid.append_report(Report("Key '" + key + "' was in rate limit."))

                        value = rate_limit.get(key)

                        try:
                            if not isinstance(value, int):
                                raise WrongTypeError(int, type(value))
                        except WrongTypeError as error:
                            are_keys_valid.append_report(Report(error))
                        else:
                            are_keys_valid.append_report(Report(f"Value of '{key}' in rate limit was an int."))

                self.report = CompositeReport(
                    is_rate_limit_key,
                    has_rate_limit_value,
                    is_rate_limit_dict,
                    are_keys_valid
                )
        else:
            self.report = CompositeReport(Report("Optional key 'rate_limit' does not exist."))

        return True


class RequestsValidator(CompositeValidator):
    """
    Validator for validating a given YAML file regarding the key 'requests'.

    Checks whether 'requests' exists in the file and the value of the key is a dict.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "requests" not in self.value:
                raise KeyNotInDictError("requests", self.value)
        except KeyNotInDictError as error:
            has_requests = Report(error)
            self.append_report(has_requests)
            return True
        else:
            has_requests = Report("Key 'requests' exists.")
            self.append_report(has_requests)
            requests = self.value.get("requests")

            try:
                if not isinstance(requests, dict):
                    raise WrongTypeError(dict, type(requests))
            except WrongTypeError as error:
                is_requests_dict = Report(error)
                self.append_report(is_requests_dict)
                return True
            else:
                is_requests_dict = Report("Value of key 'requests' is a dict.")
                self.append_report(is_requests_dict)

                for api_method in requests.values():
                    self.append_validator(ApiMethodValidator(api_method))
                return super().validate()


class ApiMethodValidator(CompositeValidator):
    """
    Validator for calling the other required Validators.
    """

    def __init__(self, value: Any):
        """
        Constructor of ApiMethodValidator.

        @param value: The value to get validated.
        """
        super().__init__(
            value,
            RequestValidator(value),
            ResponseValidator(value),
            MappingValidator(value)
        )


class RequestValidator(CompositeValidator):
    """
    Validator for validating a given API method dict regarding the key 'request'.

    Checks whether 'request' exists in the dict and the value of the key is a dict.
    Also delegation of validation of dict fields to other Validators.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def __init__(self, value: Any):
        """
        Constructor of RequestValidator.

        @param value: The value to get validated.
        """
        request: Dict[Text, Any] = value.get("request")
        super().__init__(
            value,
            # TODO: Validate whether all keys of 'template' are in 'pair_template' or 'params'
            TemplateValidator(request),
            PairTemplateValidator(request),
            ParamsValidator(request)
        )

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "request" not in self.value:
                raise KeyNotInDictError("request", self.value)
        except KeyNotInDictError as error:
            has_request = Report(error)
            self.append_report(has_request)
            return True
        else:
            has_request = Report("Key 'request' exists.")
            self.append_report(has_request)
            request = self.value.get("request")

            try:
                if not isinstance(request, dict):
                    raise WrongTypeError(dict, type(request))
            except WrongTypeError as error:
                self.append_report(Report(error))
                return True
            else:
                self.append_report(Report("Value of key 'request' is a dict."))
                return super().validate()


class TemplateValidator(Validator):
    """
    Validator for validating a given 'request' dict regarding the key 'template'.

    Checks whether 'template' exists in the 'request' and the value of the key is a String.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "template" not in self.value:
                raise KeyNotInDictError("template", self.value)
        except KeyNotInDictError as error:
            has_template = Report(error)
            self.report = has_template
        else:
            has_template = Report("Key 'template' exists.")
            template = self.value.get("template")

            try:
                if not isinstance(template, str):
                    raise WrongTypeError(str, type(self.value))
            except WrongTypeError as error:
                is_template_str = Report(error)
                self.report = is_template_str
            else:
                is_template_str = Report("template is a str.")

            self.report = CompositeReport(has_template, is_template_str)

        # TODO: Check the cases: 'template: ""' and 'template: "{currency_pair}/products"'

        return True


class PairTemplateValidator(Validator):
    """
    Validator for validating a given 'request' dict regarding the key 'pair_template'.

    Checks whether the optional key 'pair_template' exists in the file, the value of the key is None or it is a dict.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        if "pair_template" in self.value:
            has_pair_template_key = Report("Optional key 'pair_template' exists.")

            pair_template = self.value.get("pair_template")

            if pair_template is None:
                has_pair_template_value = Report("Value of optional key 'pair_template' was None.")
                self.report = CompositeReport(
                    has_pair_template_key,
                    has_pair_template_value
                )
                return True

            has_pair_template_value = Report("Value of optional key 'pair_template' was not None.")

            try:
                if not isinstance(pair_template, dict):
                    raise WrongTypeError(dict, type(pair_template))
            except WrongTypeError as error:
                is_pair_template_dict = Report(error)
            else:
                is_pair_template_dict = Report("pair_template was a dict.")

            self.report = CompositeReport(
                has_pair_template_key,
                has_pair_template_value,
                is_pair_template_dict
            )

            try:
                if "template" not in pair_template:
                    raise KeyNotInDictError("template", pair_template)
            except KeyNotInDictError as error:
                has_template_key = Report(error)
                self.report.append_report(has_template_key)
            else:
                has_template_key = Report("Key 'template' exists.")
                template_value = pair_template.get("template")

                try:
                    if not isinstance(template_value, str):
                        raise WrongTypeError(str, type(template_value))
                except WrongTypeError as error:
                    is_template_str = Report(error)
                    self.report.append_report(
                        CompositeReport(
                            has_template_key,
                            is_template_str
                        )
                    )
                else:
                    is_template_str = Report("Value of key 'template' was a str.")

                    substring_reports = CompositeReport()

                    for substring in ("{first}", "{second}"):
                        try:
                            if substring not in template_value:
                                raise SubstringNotInStringError(substring, template_value)
                        except SubstringNotInStringError as error:
                            if substring_reports:
                                substring_report = Report(f"Optional substring {substring} was not in template")
                            else:
                                substring_report = Report(error)
                        else:
                            substring_report = Report(f"Substring {substring} was in template")

                        substring_reports.append_report(substring_report)

                    self.report.append_report(
                        CompositeReport(
                            has_template_key,
                            is_template_str,
                            substring_reports
                        )
                    )
            if "lower_case" in pair_template:
                has_lower_case_key = Report("Optional key 'lower_case' exists.")
                lower_case_value = pair_template.get("lower_case")

                try:
                    if not isinstance(lower_case_value, bool):
                        raise WrongTypeError(bool, type(lower_case_value))
                except WrongTypeError as error:
                    is_lower_case_value_bool = Report(error)
                else:
                    is_lower_case_value_bool = Report("Value of key 'lower_case' is a bool.")

                self.report.append_report(
                    CompositeReport(
                        has_lower_case_key,
                        is_lower_case_value_bool
                    )
                )

            else:
                has_alias_key = Report("Optional key 'alias' does not exist.")
                self.report.append_report(has_alias_key)

            if "alias" in pair_template:
                has_alias_key = Report("Optional key 'alias' exists.")
                alias_value = pair_template.get("alias")

                try:
                    if alias_value is not None and not isinstance(alias_value, str):
                        raise WrongTypeError(str, type(alias_value))
                except WrongTypeError as error:
                    is_alias_value_str = Report(error)
                else:
                    is_alias_value_str = Report("Value of key 'alias' is a str.")

                self.report.append_report(
                    CompositeReport(
                        has_alias_key,
                        is_alias_value_str
                    )
                )

            else:
                has_alias_key = Report("Optional key 'alias' does not exist.")
                self.report.append_report(has_alias_key)

        else:
            is_pair_template_key = Report("Optional key 'pair_template' does not exist.")
            self.report = is_pair_template_key

        return True


class ParamsValidator(CompositeValidator):
    """
    Validator for validating a given 'request' dict regarding the key 'params'.

    Checks whether the optional key 'params' exists in the file, the value of the key is None or it is a dict.

    Attributes:
        value:
            The dict that shall be checked.
    """

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        if "params" in self.value:
            has_params_key = Report("Optional key 'params' does exist.")
            self.append_report(has_params_key)
            params_value = self.value.get("params")

            if params_value is None:
                is_params_value_none = Report("Value of optional key 'params' is None")
                self.append_report(is_params_value_none)

                return True

            try:
                if not isinstance(params_value, dict):
                    raise WrongTypeError(dict, type(params_value))
            except WrongTypeError as error:
                is_params_value_dict = Report(error)
                self.append_report(is_params_value_dict)
            else:
                is_params_value_dict = Report("Value of optional key 'params' is a dict.")
                self.append_report(is_params_value_dict)

                for param in params_value.values():
                    self.append_validator(ParamValidator(param))

                return super().validate()

        else:
            has_params_key = Report("Optional key 'params' does not exist.")
            self.report = has_params_key

        return True


class ParamValidator(Validator):
    """
    Validator for validating a given 'params' dict regarding a certain parameter.

    Checks whether the value of param is a dict and the key allowed or default is in the file.
    """

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if not isinstance(self.value, dict):
                raise WrongTypeError(dict, type(self.value))
        except WrongTypeError as error:
            is_value_dict = Report(error)
            self.report = is_value_dict
            return True
        else:
            is_value_dict = Report("Value of param is a dict.")
            self.report = CompositeReport(is_value_dict)

            if "allowed" in self.value:
                has_allowed_key = Report("Optional key 'allowed' is in dict.")
                allowed_value = self.value.get("allowed")

                try:
                    if not isinstance(allowed_value, dict):
                        raise WrongTypeError(list, type(allowed_value))
                except WrongTypeError as error:
                    is_allowed_value_list = Report(error)
                    self.report.append_report(
                        CompositeReport(
                            has_allowed_key,
                            is_allowed_value_list
                        )
                    )
                else:
                    is_allowed_value_list = Report("Value of key 'allowed' is a dict.")

                    try:
                        if not allowed_value:
                            raise ValueError("Value of key 'allowed' is an empty list.")
                    except ValueError as error:
                        is_allowed_list_empty = Report(error)
                    else:
                        is_allowed_list_empty = Report("Value of key 'allowed' is a non-empty dict.")

                    self.report.append_report(
                        CompositeReport(
                            has_allowed_key,
                            is_allowed_value_list,
                            is_allowed_list_empty
                        )
                    )

            else:
                has_allowed_key = Report("Optional key 'allowed' is not in dict.")
                self.report.append_report(has_allowed_key)

            if "default" in self.value:
                has_default_key = Report("Optional key 'default' is in dict.")
                default_value = self.value.get("default")

                try:
                    if default_value is None:
                        raise ValueError("Value of key 'default' is None.")
                except ValueError as error:
                    is_default_value_none = Report(error)
                else:
                    is_default_value_none = Report("Value of key 'default' is not None.")
                self.report.append_report(
                    CompositeReport(
                        has_default_key,
                        is_default_value_none
                    )
                )

            else:
                has_default_key = Report("Optional key 'default' is not in dict.")
                self.report.append_report(has_default_key)

        return True


class ResponseValidator(CompositeValidator):
    """
    Validator for validating a given API method dict regarding the key 'response'.

    Checks whether 'response' exists in the dict and the value of the key is a dict.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "response" not in self.value:
                raise KeyNotInDictError("response", self.value)
        except KeyNotInDictError as error:
            has_response = Report(error)
            self.report = has_response
        else:
            has_response = Report("Key 'response' exists.")
            response = self.value.get("response")

            try:
                if not isinstance(response, dict):
                    raise WrongTypeError(dict, type(response))
            except WrongTypeError as error:
                is_response_dict = Report(error)
                self.report = is_response_dict
            else:
                is_response_dict = Report("Value of key 'response' is a dict.")

                self.report = CompositeReport(has_response, is_response_dict)

        return True


class MappingValidator(CompositeValidator):
    """
    Validator for validating a given API method dict regarding the key 'mapping'.

    Checks whether 'mapping' exists in the dict and the value of the key is a list.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "mapping" not in self.value:
                raise KeyNotInDictError("mapping", self.value)
        except KeyNotInDictError as error:
            has_mapping_key = Report(error)
            self.append_report(has_mapping_key)
            return True
        else:
            has_mapping_key = Report("Key 'mapping' exists.")
            self.append_report(has_mapping_key)

            mapping = self.value.get("mapping")

            try:
                if not isinstance(mapping, list):
                    raise WrongTypeError(list, type(mapping))
            except WrongTypeError as error:
                is_mapping_list = Report(error)
                self.append_report(is_mapping_list)
                return True
            else:
                is_mapping_list = Report("Value of key 'mapping' was a list.")
                self.append_report(is_mapping_list)

                for mapping_entry in mapping:
                    self.append_validator(MappingEntryValidator(mapping_entry))

                return super().validate()


class MappingEntryValidator(CompositeValidator):
    """
    Validator for validating a given API method dict regarding an entry of the key 'mapping'.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def __init__(self, value: Any):
        """
        Constructor of MappingEntryValidator.

        @param value: The value to get validated.
        """
        super().__init__(
            value,
            KeyValidator(value),
            PathValidator(value),
            TypeValidator(value)
        )

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        try:
            if not isinstance(self.value, dict):
                raise WrongTypeError(dict, type(self.value))
        except WrongTypeError as error:
            is_mapping_dict = Report(error)
            self.append_report(is_mapping_dict)
            return True
        else:
            is_mapping_dict = Report("Mapping entry was a dict.")
            self.append_report(is_mapping_dict)

            return super().validate()


class KeyValidator(Validator):
    """
    Validator for validating a given 'map' dict regarding the key 'key'.

    Checks whether 'key' exists in the 'map' and the value of the key is a String.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        mapping = self.value

        try:
            if "key" not in mapping:
                raise KeyNotInDictError("key", mapping)
        except KeyNotInDictError as error:
            has_key_key = Report(error)
            self.report = has_key_key
        else:
            has_key_key = Report("Key 'key' exists.")

            key_value = mapping.get("key")

            try:
                if not isinstance(key_value, str):
                    raise WrongTypeError(str, type(key_value))
            except WrongTypeError as error:
                is_key_str = Report(error)
                self.report = is_key_str
            else:
                is_key_str = Report("Value of key 'key' was a str.")

                self.report = CompositeReport(has_key_key, is_key_str)
        return True


class PathValidator(Validator):
    """
    Validator for validating a given 'map' dict regarding the key 'path'.

    Checks whether 'path' exists in the 'map' and the value of the key is a list.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        mapping = self.value

        try:
            if "path" not in mapping:
                raise KeyNotInDictError("path", mapping)
        except KeyNotInDictError as error:
            has_path_key = Report(error)
            self.report = has_path_key
        else:
            has_path_key = Report("Key 'path' exists.")

            path_value = mapping.get("path")

            try:
                if not isinstance(path_value, list):
                    raise WrongTypeError(list, type(path_value))
            except WrongTypeError as error:
                is_path_list = Report(error)
                self.report = is_path_list
            else:
                is_path_list = Report("Value of key 'path' was a list.")

                self.report = CompositeReport(has_path_key, is_path_list)

        return True


class TypeValidator(Validator):
    """
    Validator for validating a given 'map' dict regarding the key 'type'.

    Checks whether 'type' exists in the 'map' and the value of the key is a list.

    Attributes:
        value:
            The dict that shall be checked.
    """
    value: Dict[Text, Any]

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """

        mapping = self.value

        try:
            if "type" not in mapping:
                raise KeyNotInDictError("type", mapping)
        except KeyNotInDictError as error:
            has_type_key = Report(error)
            self.report = has_type_key
        else:
            has_type_key = Report("Key 'type' exists.")

            type_value = mapping.get("type")

            try:
                if not isinstance(type_value, list):
                    raise WrongTypeError(list, type(type_value))
            except WrongTypeError as error:
                is_type_list = Report(error)
                self.report = is_type_list
            else:
                is_type_list = Report("Value of key 'type' was a list.")

                self.report = CompositeReport(has_type_key, is_type_list)

        return True


class RequestMappingValidator(Validator):
    """
    TODO: Fill out
    """
    value: Dict[Text, Any]

    @staticmethod
    def determine_table(request_name: str) -> dict:
        """
        Returns the database table based on the specified request name.

        @param request_name: The name of the request.

        @return: The database table based on the specified request name.
        """
        possible_class = {
            "currency_pairs":
                {"table": ExchangeCurrencyPair},
            "tickers":
                {"table": Ticker},
            "historic_rates":
                {"table": HistoricRate},
            "order_books":
                {"table": OrderBook},
            "trades":
                {"table": Trade}
        }
        return possible_class.get(request_name, lambda: "Invalid request class.")

    @staticmethod
    def determine_primary_keys(table_name: str) -> list:
        """
        Returns the primary keys based on the specified database table.

        @param table_name: The database table.

        @return: The primary keys based on the specified database table.
        """
        possible_primary_keys = {
            "currency_pairs":
                [],
            "tickers":
                ["time"],
            "historic_rates":
                ["time"],
            "order_books":
                ["position"],
            "trades":
                ["time"]
        }
        return possible_primary_keys.get(table_name, lambda: "Invalid request class")

    def validate(self) -> bool:
        # TODO: Philipp: Does not work, fix later
        self.report = Report("Need to be fixed later.")
        return True
        # requests = self.value
        # for request in requests.keys():
        #     table = self.determine_table(request)
        #
        #     if not isinstance(table, dict):
        #         return False
        #
        #     match_table = table['table']
        #     class_keys = [key.name for key in inspect(match_table).columns]
        #     for mapping in requests[request]['mapping']:
        #         mapping_key = mapping['key']
        #         if mapping_key != 'currency_pair_first' or mapping_key != 'currency_pair_second':
        #             try:
        #                 class_keys.index(mapping_key)
        #             except ValueError as error:
        #                 is_type_list = Report(error)
        #                 self.report = is_type_list
        #     primary_keys = self.determine_primary_keys(request)
        #     for primary_key in primary_keys:
        #         if primary_key == requests[request]['mapping']:
        #             return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore[no-untyped-def]
"""
TODO: Fill out module docstring.
"""

from typing import Dict, Tuple, Optional
from unittest.mock import Mock

import pytest

from model.database.tables import ExchangeCurrencyPair
from model.exchange.exchange import Exchange
from model.exchange.mapping import Mapping
from model.utilities.exceptions import MappingNotFoundException, DifferentExchangeContentException, \
    NoCurrencyPairProvidedException
from model.utilities.time_helper import TimeHelper


@pytest.fixture(name="exchange")
def create_exchange() -> Exchange:
    """
    TODO: Fill out
    """
    exchange_name = "test_exchange"
    exchange_dict = {
        "name": exchange_name,
        "exchange": True,
        "api_url": "https://url.to.api.com",
        "requests": []
    }
    return Exchange(exchange_dict, None, None)


class TestFormatData:
    """
    TODO: Fill out
    """

    def test_response_is_none(self, exchange):
        """Testing the handling of a response with no data in it."""
        # setup
        mappings = [Mapping("first_currency", ["data"], ["int"])]
        method = "ticker"
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        # test for ticker-data for all available currency-pairs
        response: Tuple[str, Dict[object, Optional[Dict]]] = (exchange.name, {None: None})
        result = exchange.format_data(method, response, start_time, time)

        with pytest.raises(StopIteration):
            next(result)

        # test for multiple currency-pairs
        first_ecp_mock = Mock(spec=ExchangeCurrencyPair)
        second_ecp_mock = Mock(spec=ExchangeCurrencyPair)
        third_ecp_mock = Mock(spec=ExchangeCurrencyPair)
        response = (exchange.name, {first_ecp_mock: None, second_ecp_mock: None, third_ecp_mock: None})
        result = exchange.format_data(method, response, start_time, time)

        with pytest.raises(StopIteration):
            next(result)

    def test_empty_response(self, exchange):
        """Testing the handling of an empty response."""
        # todo: test for pairs and for all together
        # setup
        mappings = [Mapping("first_currency", ["data"], ["int"])]
        method = "ticker"
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        # test for ticker-data for all available currency-pairs
        response: Tuple[str, Dict[object, Optional[Dict]]] = (exchange.name, {None: {}})
        result = exchange.format_data(method, response, start_time, time)

        with pytest.raises(StopIteration):
            next(result)

        # test for multiple currency-pairs
        first_ecp_mock = Mock(spec=ExchangeCurrencyPair)
        second_ecp_mock = Mock(spec=ExchangeCurrencyPair)
        third_ecp_mock = Mock(spec=ExchangeCurrencyPair)
        response = (exchange.name, {first_ecp_mock: {}, second_ecp_mock: {}, third_ecp_mock: {}})

        result = exchange.format_data(method, response, start_time, time)

        with pytest.raises(StopIteration):
            next(result)

    def test_no_mapping_available(self, exchange):
        """Testing the case if the data is valid but there is no mapping available."""
        # Method is there but no mappings behind
        mappings = []
        method: str = "ticker"
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        response: Tuple[str, Dict[object, Optional[Dict]]] = (exchange.name, {None: {"response": "Hello"}})
        with pytest.raises(MappingNotFoundException):
            next(exchange.format_data(method, response, start_time, time))

        # No method available
        exchange.response_mappings = {}
        with pytest.raises(MappingNotFoundException):
            next(exchange.format_data(method, response, start_time, time))

    def test_response_from_diff_exchange(self, exchange):
        """Testing the case where the given response-dict is from a different exchange.
           This is detected by the given name."""
        mappings = []
        method: str = "ticker"
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        response: Tuple[str, Dict[object, Optional[Dict]]] = ("other exchange", {None: {"response": "Hello"}})
        with pytest.raises(DifferentExchangeContentException):
            next(exchange.format_data(method, response, start_time, time))

    def test_all_request_but_no_cp_first_or_second(self, exchange):
        """ Testing special case where the response contains all the available data but
            there is no currency_pair_first or currency_pair_second as name for a mapping."""

        method: str = "ticker"
        # currency_pair_first missing
        mappings = [Mapping("currency_pair_second", ["second"], ["str"]),
                    Mapping("value", ["value"], ["str", "int"])]
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        # noinspection PyTypeChecker
        response: Tuple[str, Dict[object, Optional[Dict]]] = (
            exchange.name, {None: [{"first": "btc", "second": "eth", "value": 1},
                                   {"first": "eth", "second": "xrp", "value": 2},
                                   {"first": "btc", "second": "usd", "value": 3},
                                   {"first": "eth", "second": "usdt", "value": 4}]})
        with pytest.raises(NoCurrencyPairProvidedException):
            next(exchange.format_data(method, response, start_time, time))

        # currency_pair_second missing
        mappings = [Mapping("currency_pair_first", ["first"], ["str"]),
                    Mapping("value", ["value"], ["str", "int"])]
        exchange.response_mappings = {method: mappings}
        with pytest.raises(NoCurrencyPairProvidedException):
            next(exchange.format_data(method, response, start_time, time))

        # both missing
        mappings = [Mapping("value", ["value"], ["str", "int"])]
        exchange.response_mappings = {method: mappings}
        with pytest.raises(NoCurrencyPairProvidedException):
            next(exchange.format_data(method, response, start_time, time))

    def test_all_request_guarding_cp(self):
        """ Testing format_data for a response that coontains all the available data.
            The data for each currency pair is guarded by the formatted string from each pair."""
        # todo: currently there is no way to get to this information lol

    def test_all_request_list_dict(self, exchange):
        """ Test for a request which contains all the available data.
            The mocked response is a list of dictionaries which contain the data that
            is to be formatted."""

        mappings = [Mapping("currency_pair_first", ["first"], ["str"]),
                    Mapping("currency_pair_second", ["second"], ["str"]),
                    Mapping("value", ["value"], ["str", "int"])]
        method: str = "ticker"
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        # noinspection PyTypeChecker
        response: Tuple[str, Dict[object, Optional[Dict]]] = (
            exchange.name, {None: [{"first": "btc", "second": "eth", "value": 1},
                                   {"first": "eth", "second": "xrp", "value": 2},
                                   {"first": "btc", "second": "usd", "value": 3},
                                   {"first": "eth", "second": "usdt", "value": 4}]})

        # result, keys = exchange.format_data(method, response, start_time, time)
        result = exchange.format_data(method, response, start_time, time)

        value_list = [(start_time, time, "btc", "eth", 1),
                      (start_time, time, "eth", "xrp", 2),
                      (start_time, time, "btc", "usd", 3),
                      (start_time, time, "eth", "usdt", 4)]
        key_list = ["start_time", "time", "currency_pair_first", "currency_pair_second", "value"]

        for got in result:
            assert value_list == got[0]
            assert key_list == got[1]

    def test_all_request_list_dict_dict(self, exchange):
        """Test for a request which contains all the available data."""
        mappings = [Mapping("currency_pair_first", ["data", "first"], ["str"]),
                    Mapping("currency_pair_second", ["data", "second"], ["str"]),
                    Mapping("value", ["data", "value"], ["str", "int"])]
        method: str = "ticker"
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        # noinspection PyTypeChecker
        response: Tuple[str, Dict[object, Optional[Dict]]] = (
            exchange.name, {None: [{"data": {"first": "btc", "second": "eth", "value": 5}},
                                   {"data": {"first": "eth", "second": "xrp", "value": 6}},
                                   {"data": {"first": "btc", "second": "usd", "value": 7}},
                                   {"data": {"first": "eth", "second": "usdt", "value": 8}}]})

        result = exchange.format_data(method, response, start_time, time)

        value_list = [(start_time, time, "btc", "eth", 5),
                      (start_time, time, "eth", "xrp", 6),
                      (start_time, time, "btc", "usd", 7),
                      (start_time, time, "eth", "usdt", 8)]
        key_list = ["start_time", "time", "currency_pair_first", "currency_pair_second", "value"]

        for got in result:
            assert value_list == got[0]
            assert key_list == got[1]

    def test_all_request_dict_list_dict(self, exchange):
        """Test for a request which contains all the available data."""

        mappings = [Mapping("currency_pair_first", ["data", "first"], ["str"]),
                    Mapping("currency_pair_second", ["data", "second"], ["str"]),
                    Mapping("value", ["data", "value"], ["str", "int"])]
        method: str = "ticker"
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        # noinspection PyTypeChecker
        response: Tuple[str, Dict[object, Optional[Dict]]] = (
            exchange.name, {None: {"data": [{"first": "btc", "second": "eth", "value": 1},
                                            {"first": "eth", "second": "xrp", "value": 2},
                                            {"first": "btc", "second": "usd", "value": 3},
                                            {"first": "eth", "second": "usdt", "value": 4}]}})

        result = exchange.format_data(method, response, start_time, time)

        value_list = [(start_time, time, "btc", "eth", 1),
                      (start_time, time, "eth", "xrp", 2),
                      (start_time, time, "btc", "usd", 3),
                      (start_time, time, "eth", "usdt", 4)]
        key_list = ["start_time", "time", "currency_pair_first", "currency_pair_second", "value"]

        for got in result:
            assert value_list == got[0]
            assert key_list == got[1]

    def test_cp_request_dict(self, exchange):
        """ Testing the formatting of individual responses for each currency pair.
            The data is contained in a dictionary."""
        exchange.request_urls = {'ticker': {'pair_template': {'template': '{first}_{second}', 'lower_case': False}}}
        mappings = [Mapping('value1', ['v1'], ['str', 'int']),
                    Mapping('value2', ['v2'], ['str']),
                    Mapping('value3', ['v3'], ['str'])]
        method: str = 'ticker'
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        cp1_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp1_mock.id = 1
        cp2_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp2_mock.id = 2
        cp3_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp3_mock.id = 3
        cp4_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp4_mock.id = 4

        # noinspection PyTypeChecker
        response: Tuple[str, Dict[object, Optional[Dict]]] = (
            exchange.name, {cp1_mock: {'v1': '10', 'v2': 'a', 'v3': 'b'},
                            cp2_mock: {'v1': '11', 'v2': 'c', 'v3': 'd'},
                            cp3_mock: {'v1': '12', 'v2': 'e', 'v3': 'f'},
                            cp4_mock: {'v1': '13', 'v2': 'g', 'v3': 'h'}})
        value_list = [(start_time, time, 10, 'a', 'b', 1),
                      (start_time, time, 11, 'c', 'd', 2),
                      (start_time, time, 12, 'e', 'f', 3),
                      (start_time, time, 13, 'g', 'h', 4)]
        key_list = ['start_time', 'time', 'value1', 'value2', 'value3', 'exchange_pair_id']

        result = exchange.format_data(method, response, start_time, time)

        for i, got in enumerate(result):
            assert value_list[i] == got[0][0]
            assert key_list == got[1]

    def test_cp_request_dict_dict(self, exchange):
        """ Testing the formatting of individual responses for each currency pair.
            The data is guarded by a dict before accessing the actual data-dict."""
        exchange.request_urls = {'ticker': {'pair_template': {'template': '{first}_{second}', 'lower_case': False}}}
        mappings = [Mapping('value1', ['data', 'v1'], ['str', 'int']),
                    Mapping('value2', ['data', 'v2'], ['str']),
                    Mapping('value3', ['data', 'v3'], ['str'])]
        method: str = 'ticker'
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        cp1_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp1_mock.id = 1
        cp1_mock.first.name = 'btc'
        cp1_mock.first.name = 'eth'
        cp2_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp2_mock.id = 2
        cp1_mock.first.name = 'eth'
        cp1_mock.first.name = 'xrp'
        cp3_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp3_mock.id = 3
        cp1_mock.first.name = 'btc'
        cp1_mock.first.name = 'usd'
        cp4_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp4_mock.id = 4
        cp1_mock.first.name = 'btc'
        cp1_mock.first.name = 'usdt'

        # noinspection PyTypeChecker
        response: Tuple[str, Dict[object, Optional[Dict]]] = (
            exchange.name, {cp1_mock: {'data': {'v1': '10', 'v2': 'a', 'v3': 'b'}},
                            cp2_mock: {'data': {'v1': '11', 'v2': 'c', 'v3': 'd'}},
                            cp3_mock: {'data': {'v1': '12', 'v2': 'e', 'v3': 'f'}},
                            cp4_mock: {'data': {'v1': '13', 'v2': 'g', 'v3': 'h'}}})
        value_list = [(start_time, time, 10, 'a', 'b', 1),
                      (start_time, time, 11, 'c', 'd', 2),
                      (start_time, time, 12, 'e', 'f', 3),
                      (start_time, time, 13, 'g', 'h', 4)]
        key_list = ['start_time', 'time', 'value1', 'value2', 'value3', 'exchange_pair_id']

        result = exchange.format_data(method, response, start_time, time)

        for i, got in enumerate(result):
            assert value_list[i] == got[0][0]
            assert key_list == got[1]

    def test_cp_request_dict_cp_guard_dict(self, exchange):
        """ Testing the formatting of individual responses for each currency pair.
            The desired data is guarded by the formatted currency pair string."""
        exchange.request_urls = {'ticker': {'pair_template': {'template': '{first}_{second}', 'lower_case': False}}}
        mappings = [Mapping('value1', ['data', 'v1'], ['str', 'int']),
                    Mapping('value2', ['data', 'v2'], ['str']),
                    Mapping('value3', ['data', 'v3'], ['str'])]
        method: str = 'ticker'
        exchange.response_mappings = {method: mappings}
        start_time = TimeHelper.now()
        time = TimeHelper.now()

        cp1_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp1_mock.id = 1
        cp2_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp2_mock.id = 2
        cp3_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp3_mock.id = 3
        cp4_mock: ExchangeCurrencyPair = Mock(spec=ExchangeCurrencyPair)
        cp4_mock.id = 4

        # noinspection PyTypeChecker
        response: Tuple[str, Dict[object, Optional[Dict]]] = (
            exchange.name, {cp1_mock: {'data': {'v1': '10', 'v2': 'a', 'v3': 'b'}},
                            cp2_mock: {'data': {'v1': '11', 'v2': 'c', 'v3': 'd'}},
                            cp3_mock: {'data': {'v1': '12', 'v2': 'e', 'v3': 'f'}},
                            cp4_mock: {'data': {'v1': '13', 'v2': 'g', 'v3': 'h'}}})
        value_list = [(start_time, time, 10, 'a', 'b', 1),
                      (start_time, time, 11, 'c', 'd', 2),
                      (start_time, time, 12, 'e', 'f', 3),
                      (start_time, time, 13, 'g', 'h', 4)]
        key_list = ['start_time', 'time', 'value1', 'value2', 'value3', 'exchange_pair_id']

        result = exchange.format_data(method, response, start_time, time)

        for i, got in enumerate(result):
            assert value_list[i] == got[0][0]
            assert key_list == got[1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module to schedule the program. The scheduler receives a job list, validates it and calls the methods
to request, extract and persist data asynchronously. The scheduler is programmed to request several currency-pairs
at once, however it can be changed to an vertical requesting (i.e. on currency-pair at a time). This may be useful
for a high amount of currency-pairs, to avoid filling the RAM.
"""

import asyncio
import logging
from asyncio import Future
from typing import Callable, Any, Optional, Union, Coroutine, List, Dict, Tuple

from model.database.db_handler import DatabaseHandler
from model.database.tables import Ticker, Trade, OrderBook, HistoricRate, ExchangeCurrencyPair, DatabaseTable
from model.exchange.exchange import Exchange
from model.scheduling.job import Job
from model.utilities.exceptions import MappingNotFoundException
from model.utilities.kill_switch import KillSwitch
from model.utilities.loading_bar import Loader
from model.utilities.time_helper import TimeHelper


class Scheduler:
    """
    The scheduler is in charge of requesting, filtering and persisting data received from the exchanges.
    Every x minutes the scheduler will be called to run the jobs created by the user in config.yaml.
    Attributes like frequency or job_list can also be set by the user in config.yaml.
    """

    def __init__(self, database_handler: DatabaseHandler, job_list: List[Job],
                 asynchronicity: Union[int, bool], frequency: Union[str, int, float]):
        """
        Initializer for a Scheduler.

        @param database_handler: Handler that is called for everything that needs information from the database.
        @type database_handler: DatabaseHandler
        @param job_list: List of Jobs. A job can be created with the specific yaml-template in config.yaml.
        @type job_list: list[Job]
        @param asynchronicity: Specifies the requesting method, horizontal or vertical.
        @param frequency: The interval in minutes with that the run() method gets called.
        @type frequency: Any
        """
        self.database_handler = database_handler
        self.job_list = job_list
        self.asynchronicity = asynchronicity
        self.frequency = frequency * 60 if isinstance(frequency, (int, float)) else frequency
        self._validated = False

    async def start(self) -> None:
        """
        Starts the process of requesting, filtering and persisting data for each job every x minutes.
        If a job takes longer than the frequency. The scheduler will wait until the job is finished
        and then start the jobs immediately again.
        Otherwise the scheduler will wait x minutes until it starts the jobs again.
        The interval begins counting down at the start of the current iteration.
        """
        runs: list[Union[Coroutine[Any, Any, None], Future[Any]]] = [self.run(job) for job in self.job_list]

        if isinstance(self.frequency, (int, float)):
            runs.append(asyncio.sleep(self.frequency))
        await asyncio.gather(*runs)

    async def run(self, job: Job) -> None:
        """
        The method represents one execution of the given job.

        @param job: The job that will be executed.
        @type job: Job
        """

        if not self._validated:
            await self.validate_job()

        request = self.determine_task(job.request_name)
        request_fun = request.get("function")
        request_table = request.get("table")

        if self.asynchronicity is False:
            for exchange, currency_pairs in job.exchanges_with_pairs.items():
                for currency_pair, last_row_id in currency_pairs.items():
                    continue_run = True
                    while continue_run:
                        if KillSwitch().stay_alive is False:
                            print("\nTask got terminated.")
                            logging.info("Task got terminated.")
                            break
                        last_row_id = job.exchanges_with_pairs.get(exchange).get(currency_pair)
                        continue_run, job.exchanges_with_pairs = \
                            await request_fun(request_table, {exchange: {currency_pair: last_row_id}})
                        if not continue_run:
                            continue

        else:
            continue_run = True
            while continue_run:
                if KillSwitch().stay_alive is False:
                    print("\nTask got terminated.")
                    logging.info("Task got terminated.")
                    break
                continue_run, job.exchanges_with_pairs = await request_fun(request_table, job.exchanges_with_pairs)

    def determine_task(self, request_name: str) -> Dict[str, Union[Callable[..., None]]]:
        """
        Returns the method that is to execute based on the given request name.

        @param request_name: Name of the request.
        @type request_name: str

        @return: Method for the request name or a string that the request is false.
        @rtype: dict[str, Callable]
        """

        possible_requests = {
            "currency_pairs":
                {"function": self.get_currency_pairs,
                 "table": ExchangeCurrencyPair},
            "tickers":
                {"function": self.request_format_persist,
                 "table": Ticker},
            "historic_rates":
                {"function": self.request_format_persist,
                 "table": HistoricRate},
            "order_books":
                {"function": self.request_format_persist,
                 "table": OrderBook},
            "trades":
                {"function": self.request_format_persist,
                 "table": Trade}
        }

        return possible_requests.get(request_name, {
            "function": lambda: "Invalid request name.",
            "table": None
        })

    async def validate_job(self) -> None:
        """
        This methods validates the job_list given to the scheduler instance. If the job-list does not contain
        any "exchange_with_pairs" or no currency_pair for an exchange, the job is removed from the list.
        This happens of the user specified currency-pairs in the config but an exchange does not offer that pair.

        @return: New job_list without empty job and sets self.validated: True if the validation is successful.
        """

        self.job_list = await self.get_currency_pairs(self.job_list)
        self.remove_invalid_jobs(self.job_list)
        if self.job_list:
            self._validated = True

    def remove_invalid_jobs(self, jobs: List[Job]) -> List[Job]:
        """
        Method to clean up the job list. If the job list is empty, shut down program.
        Else the algorithm will go through every job specification and delete empty jobs or exchanges.

        @param jobs: List of all jobs specified by the config
        @type jobs: list[Job]
        @return: List of jobs, cleaned by empty or invalid jobs
        @rtype: list[Job]
        """
        if not jobs:
            logging.error("No or invalid Job(s).")

            print("\nNo currency-pair(s) found for the specified exchange(s). "
                  "Please check your request settings in the configuration file.")
            raise SystemExit

        for job in jobs:
            if job.request_name == "currency_pairs":
                print("\nDone loading Currency-Pairs.")
                raise SystemExit

            if job.exchanges_with_pairs:
                for exchange in job.exchanges_with_pairs.copy():
                    # Delete exchanges with no API for that request type
                    if job.request_name not in list(exchange.file["requests"].keys()):
                        job.exchanges_with_pairs.pop(exchange)
                        logging.info("%s has no %s request method and was removed.",
                                     exchange.name.capitalize(), job.request_name)
                    # Delete exchanges with no matching Currency_Pairs
                    elif not job.exchanges_with_pairs[exchange]:
                        job.exchanges_with_pairs.pop(exchange)
                        logging.info("%s has no matching currency_pairs.", exchange.name.capitalize())

                # Delete empty jobs, if the previous conditions removed all exchanges
                if not job.exchanges_with_pairs:
                    jobs.remove(job)

            else:
                # remove job if initially empty
                jobs.remove(job)

        if jobs:
            # If there are jobs left, return them
            for job in jobs:
                print(f"Requesting {len(job.exchanges_with_pairs.keys())} exchange(s) for job: {job.name}.")
            return jobs
        # Reenter the method to get into the first else (down) condition and shut down process
        self.remove_invalid_jobs(jobs)

    async def update_currency_pairs(self, ex: Exchange) -> List[None]:
        """
        This method requests the currency_pairs.

        @param ex: Current exchange object.
        @type ex: Exchange

        @return: Empty list if no response from the exchange.
        @rtype: list
        """
        response = await ex.request_currency_pairs()
        if response[1]:
            try:
                formatted_response = ex.format_currency_pairs(response)
                self.database_handler.persist_exchange_currency_pairs(formatted_response,
                                                                      is_exchange=ex.is_exchange)
            except (MappingNotFoundException, TypeError, KeyError):
                logging.exception("Error updating currency_pairs for %s", ex.name.capitalize())
                return []
        else:
            return []

    async def get_currency_pairs(self, job_list: List[Job]) -> List[Job]:
        """
        Method to get all exchange currency pairs. First the database is queried, if the result is [], the exchanges
        api for all currency pairs is called.

        @param job_list: list of all jobs, including Job-Objects.
        @type job_list: list[Job]

        @return job_list with updated exchange_currency_pairs.
        @rtype: list[Job]
        """
        loader: Loader

        for job in job_list:
            job_params = job.job_params
            exchanges = list(job.exchanges_with_pairs.keys())

            logging.info("Loading and/or updating exchange currency pairs..")

            with Loader("Checking exchange currency-pairs...", "", max_counter=len(exchanges)) as loader:
                exchanges_to_update = list()
                for exchange in exchanges:
                    if job_params["update_cp"] or job.request_name == "currency_pairs" or \
                            not self.database_handler.get_all_currency_pairs_from_exchange(exchange.name):
                        exchanges_to_update.append(self.update_currency_pairs(exchange))
                    loader.increment()

            with Loader("Requesting exchange currency-pairs...", "", max_counter=len(exchanges_to_update)) as loader:
                for exchange in exchanges_to_update:
                    await exchange
                    loader.increment()

            with Loader("Loading exchange currency-pairs...", "", max_counter=len(exchanges)) as loader:
                for exchange in exchanges:
                    if job.request_name != "currency_pairs":
                        job.exchanges_with_pairs[exchange] = dict.fromkeys(
                            self.database_handler.get_exchanges_currency_pairs(
                                exchange.name,
                                job_params["currency_pairs"],
                                job_params["first_currencies"],
                                job_params["second_currencies"]
                            ))
                    loader.increment()

        return job_list

    async def request_format_persist(self,
                                     request_table: DatabaseTable,
                                     exchanges_with_pairs: Dict[Exchange, Dict[ExchangeCurrencyPair, None]]) \
            -> Tuple[bool, Dict[Exchange, Dict[ExchangeCurrencyPair, None]]]:
        """"
        Gets the job done. The request are sent concurrently and awaited. Afterwards the responses
        are formatted via "found_exchange.format_data()", a method from the Exchange Class. The formatted
        responses and the mappings (IMPORTANT: THE ORDER OF THE RESPONSE TUPLES AND MAPPINGS REMAIN UNTOUCHED)
        are given to the DatabaseHandler where they are checked, single items removed (who do not belong in the
        database table, especially the start_time is only present in the ticker table) and persisted.

        This method works for all kind of request, except the currency_pairs. To add a new request-type
        (i.e. like order_books, ticker,..) add a new item into 'possible_requests' from self.determine_task(),
        create a new database class (i.e. like OrderBook, Ticker,...) and expand the yaml-file for each exchange.

        Please ensure the following:
            - The database columns MUST match the mapping-keys from the yaml-file.
            - The order of the mapping-keys from the yaml-file does not matter. It is matched to the values
                in "Exchange.format_data()" and handed over to SQLAlchemy (where a new object is created for each row)
                via **kwargs.
            - The DatabaseHandler will reject to persist new items if any primary key is emtpy.
            - For more detailed instructions, including an example, see into the handbook.

        @param request_table: The database table storing the data.
        @type request_table: object
        @param exchanges_with_pairs: The exchanges including currency pairs to be queried. The value of the dict
                                    contains the last row_id of the last insert. Useful to retrieve the next timestamp
                                    for HR requests.
        @type exchanges_with_pairs: dict[Exchange, Dict[ExchangeCurrencyPair, Optional[int]]

        @return: A bool whether the job has to run again and a list of updated exchanges.
        @rtype: tuple[bool, dict[Exchange, list[ExchangeCurrencyPair]]]
        """
        table_name = request_table.__tablename__.capitalize()

        logging.info("Starting to request %s.", table_name)

        start_time = TimeHelper.now()

        total = sum([len(v) for v in exchanges_with_pairs.values()])

        loader: Loader
        with Loader("Requesting data...", "", max_counter=total) as loader:
            responses = await asyncio.gather(
                *(ex.request(request_table, exchanges_with_pairs[ex], loader=loader) for ex in
                  exchanges_with_pairs.keys())
            )

        counter = {}

        for response in responses:
            if not response:
                continue

            response_time = response[0]
            exchange_name = response[1]
            found_exchange: Optional[Exchange] = None

            for exchange in exchanges_with_pairs.keys():
                # finding the right exchange object
                if exchange.name.upper() == exchange_name.upper():
                    found_exchange = exchange
                    break

            if found_exchange:
                try:
                    formatted_response = found_exchange.format_data(request_table.__tablename__,
                                                                    response[1:],
                                                                    start_time=start_time,
                                                                    time=response_time)

                    if formatted_response:
                        counter[found_exchange] = self.database_handler.persist_response(exchanges_with_pairs,
                                                                                         found_exchange,
                                                                                         request_table,
                                                                                         formatted_response)

                except (MappingNotFoundException, TypeError, KeyError):
                    logging.exception("Exception formatting or persisting data for %s", found_exchange.name)
                    continue

        if request_table.__name__ == "HistoricRate":
            updated_job: Dict[Exchange, Any] = {}
            for exchange, value in counter.items():
                if value:
                    updated_job[exchange] = value

            if updated_job:
                return True, updated_job

        logging.info("Done collecting %s.", table_name)

        return False, exchanges_with_pairs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module for interacting with the database.

Classes:
    - DatabaseHandler:
        Methods:
            - Constructor
            - Session-Factory
            - Get methods
            - Persist methods
"""

import importlib
import logging
import os
from contextlib import contextmanager
from datetime import datetime, timedelta
from itertools import product
from typing import List, Iterable, Optional, Generator, Any, Iterator, Dict, Tuple, Union

import sqlalchemy.orm
from pandas import DataFrame
from pandas import read_sql_query as pd_read_sql_query
from sqlalchemy import create_engine, MetaData, tuple_, func, inspect
from sqlalchemy.exc import ProgrammingError, OperationalError, SQLAlchemyError
from sqlalchemy.orm import sessionmaker, Session, Query, aliased
from sqlalchemy_utils import database_exists, create_database

from model.database.tables import ExchangeCurrencyPair, Exchange, Currency, DatabaseTable
from model.utilities.time_helper import TimeHelper, TimeUnit
from model.utilities.utilities import split_str_to_list, install_packages


class DatabaseHandler:
    """
    Class which handles every interaction with the database.
    This includes most of the time checking if values exist in
    the database or storing/querying values.

    For querying and storing values the library sqlalchemy is used.

    Attributes:
        session_factory: sessionmaker
           Factory for connections to the database.
    """

    def __init__(
            self,
            metadata: MetaData,
            sqltype: str,
            client: str,
            user_name: str,
            password: str,
            host: str,
            port: str,
            db_name: str,
            min_return_tuples: int = 1,
            path: Optional[str] = None,
            debug: bool = False):
        """
        Initializes the database-handler.

        Builds the connection-string and tries to connect to the database.

        Creates with the given metadata tables which do not already exist.
        Won't make a new table if the name already exists,
        so changes to the table-structure have to be made by hand in the database
        or the table has to be deleted.

        Initializes the sessionFactory with the created engine.
        Engine variable is no attribute and currently only exists in the constructor.

        @param metadata: Metadata Information about the table-structure of the database.
                        See tables.py for more information.
        @type metadata: MetaData
        @param sqltype: Type of the database sql-dialect. ('postgresql, mariadb, mysql, sqlite' for us)
        @type sqltype: str
        @param client: Name of the Client which is used to connect to the database.
        @type client: str
        @param user_name: Username under which this program connects to the database.
        @type user_name: str
        @param password: Password for this username.
        @type password: str
        @param host: Hostname or host-address from the database.
        @type host: str
        @param port: Connection-Port (usually 5432 for Postgres)
        @type port: str
        @param db_name: Name of the database.
        @type db_name: str
        @param min_return_tuples: Minimum amount of tuples returned in order to keep exchange alive
        @type min_return_tuples: int
        @param path: Path to the database directory.
        @type path: str
        @param debug: Indicates if the debug mode is on.
        @type debug: bool
        """
        if not path:
            path = os.getcwd()

        conn_strings = {"debug": "sqlite://",
                        "sqlite": f"{sqltype}:///{path}/{db_name}.db",
                        "postgresql": f"{sqltype}+{client}://{user_name}:{password}@{host}:{port}/{db_name}",
                        "mariadb": f"{sqltype}+{client}://{user_name}:{password}@{host}:{port}/{db_name}",
                        "mysql": f"{sqltype}+{client}://{user_name}:{password}@{host}:{port}/{db_name}"}
        if debug:
            conn_string = conn_strings["debug"]
        else:
            conn_string = conn_strings[sqltype]

        logging.info("Connection String is: %s", conn_string)

        try:
            engine = create_engine(conn_string)
        except ModuleNotFoundError:
            install_packages(client)
            engine = create_engine(conn_string)

        if not database_exists(engine.url):
            create_database(engine.url)
            print(f"Database '{db_name}' created", end="\n\n")
            logging.info("Database '%s' created", db_name)

        try:  # this is done since one can't test if view-table exists already. if it does an error occurs
            metadata.create_all(engine)
        except (ProgrammingError, OperationalError):
            message = "Database Views already exist. If you need to alter or recreate tables delete all views manually."
            logging.info(message)

        self.session_factory: sessionmaker = sessionmaker(bind=engine)
        self._min_return_tuples = min_return_tuples

        if sqltype == 'mariadb':
            sqltype = "mysql"
        self.insert_module = importlib.import_module(f"sqlalchemy.dialects.{sqltype}")


    @contextmanager
    def session_scope(self) -> Generator[Session, None, None]:
        """Provide a transactional scope around a series of operations."""
        session = self.session_factory()
        try:
            yield session
            session.commit()
        except SQLAlchemyError as ex:
            #  Postgresql throw integrity errors which where not caught.
            #  Sqlite on the other hand not. For reproducibility: B2bx, BTC-USD
            logging.exception(ex)
            session.rollback()
        finally:
            session.close()

    def get_currency_id(self, currency_name: str) -> Optional[int]:
        """
        Gets the ID of the given currency if it exists in the database.

        @param currency_name: The name of the currency.

        @return: The ID of the given currency or None if no currency with the given name exists in the database.
        """
        with self.session_scope() as session:
            return session.query(Currency.id).filter(Currency.name == currency_name.upper()).scalar()

    def get_exchange_id(self, exchange_name: str) -> int:
        """
        Gets the ID of the given exchange if it exists in the database.

        @param exchange_name: The name of the exchange.

        @return: The ID of the given exchange or None if no exchange with the given name exists in the database.
        """
        with self.session_scope() as session:
            return session.query(Exchange.id).filter(Exchange.name == exchange_name.upper()).scalar()

    def get_currency_pairs(self, exchange_name: str, currency_pairs: List[Dict[str, str]]) \
            -> List[ExchangeCurrencyPair]:
        """
        Returns all ExchangeCurrencyPairs for the given exchange if they fit any
        currency pairs in the given list of dictionaries.

        @param exchange_name: Name of the exchange.
        @type exchange_name: str
        @param currency_pairs: List of the currency pairs that should be found.
                               Each dictionary should contain the keys 'first' and 'second'
                               which contain the names of the currencies.
        @type currency_pairs: list[dict[str, str]]

        @return: List of all found currency pairs on this exchange based on the given pair combinations.
        @rtype: list[ExchangeCurrencyPair]
        """
        found_currency_pairs: List[ExchangeCurrencyPair] = list()

        if exchange_name:
            exchange_id: int = self.get_exchange_id(exchange_name)
            with self.session_scope() as session:
                if currency_pairs is not None:
                    for currency_pair in currency_pairs:
                        first_currency = currency_pair["first"]
                        second_currency = currency_pair["second"]
                        if first_currency and second_currency:
                            first_id: int = self.get_currency_id(first_currency)
                            second_id: int = self.get_currency_id(second_currency)

                            found_currency_pair = session.query(ExchangeCurrencyPair).filter(
                                ExchangeCurrencyPair.exchange_id == exchange_id,
                                ExchangeCurrencyPair.first_id == first_id,
                                ExchangeCurrencyPair.second_id == second_id).first()

                            if found_currency_pair is not None:
                                found_currency_pairs.append(found_currency_pair)
                    session.expunge_all()

        return found_currency_pairs

    @staticmethod
    def _get_exchange_currency_pair(session: sqlalchemy.orm.Session,
                                    exchange_name: str,
                                    first_currency_name: str,
                                    second_currency_name: str) -> Optional[ExchangeCurrencyPair]:
        """
        Checks if there is a currency pair in the database with the given parameters and
        returns it if so.

        @param session: Session from the session-factory
        @type: session: sqlalchemy.orm.Session
        @param exchange_name: Name of the exchange.
        @type exchange_name: str
        @param first_currency_name: Name of the first currency in the currency pair.
        @type first_currency_name: str
        @param second_currency_name: Name of the second currency in the currency pair.
        @type second_currency_name: str
        @return: The ExchangeCurrencyPair which fulfills all the requirements or None
                 if no such ExchangeCurrencyPair exists.
        @rtype: Optional[ExchangeCurrencyPair]
        """
        if exchange_name is None or first_currency_name is None or second_currency_name is None:
            return None

        ex = session.query(Exchange).filter(Exchange.name == exchange_name.upper()).first()
        first = session.query(Currency).filter(Currency.name == first_currency_name.upper()).first()
        second = session.query(Currency).filter(Currency.name == second_currency_name.upper()).first()

        return session.query(ExchangeCurrencyPair).filter(
            ExchangeCurrencyPair.exchange == ex,
            ExchangeCurrencyPair.first == first,
            ExchangeCurrencyPair.second == second
        ).first()

    def get_exchanges_currency_pairs(self,
                                     exchange_name: str,
                                     currency_pairs: List[Dict[str, str]],
                                     first_currencies: Union[List[str], str],
                                     second_currencies: Union[List[str], str]) -> List[ExchangeCurrencyPair]:

        """
        Collects and returns all currency pairs for the given exchange that either have any
        of the currencies of first_currencies/second_currencies as a currency as
        first/second or match a specific pair in currency_pairs.

        @param exchange_name: str
            Name of the exchange.
        @param currency_pairs: List[Dict[str, str]]
            List of specific currency pairs that should be found.
            Dictionary should have the following keys:
                first: 'name of the first currency'
                second: 'name of the second currency'
        @param first_currencies: List[str]
            List of currency names that are viable as first currency.
            All pairs that have any of the given names as first currency will be returned.
        @param second_currencies: List[str]
            List of currency names that are viable as second currency.
            All pairs that have any of the given names as second currency will be returned.
        @return:
            All ExchangeCurrencyPairs of the given Exchange that fulfill any
            of the above stated conditions.
        """
        if isinstance(first_currencies, str):
            first_currencies = split_str_to_list(first_currencies)

        if isinstance(second_currencies, str):
            second_currencies = split_str_to_list(second_currencies)

        found_currency_pairs: List[ExchangeCurrencyPair] = list()

        if currency_pairs:

            if "all" in currency_pairs:
                found_currency_pairs.extend(self.get_all_currency_pairs_from_exchange(exchange_name))
            elif isinstance(currency_pairs, str):
                currency_pairs = split_str_to_list(currency_pairs)
                currency_pairs = [{"first": split_str_to_list(pair, "-")[0],
                                   "second": split_str_to_list(pair, "-")[-1]} for pair in currency_pairs]
                found_currency_pairs.extend(self.get_currency_pairs(exchange_name, currency_pairs))

        if first_currencies and second_currencies:
            currency_pairs = list(product(first_currencies, second_currencies))
            currency_pairs = [{"first": pair[0], "second": pair[1]} for pair in currency_pairs]
            found_currency_pairs.extend(self.get_currency_pairs(exchange_name, currency_pairs))

        elif first_currencies or second_currencies:
            found_currency_pairs.extend(self.get_currency_pairs_with_first_currency(exchange_name, first_currencies))
            found_currency_pairs.extend(self.get_currency_pairs_with_second_currency(exchange_name, second_currencies))

        result: list[ExchangeCurrencyPair] = list()

        for pair in found_currency_pairs:
            if not any(pair.id == result_pair.id for result_pair in result):
                result.append(pair)
        return result

    def get_all_currency_pairs_from_exchange(self, exchange_name: str) -> List[ExchangeCurrencyPair]:
        """
        @param exchange_name: Name of the exchange that the currency-pairs should be queried for.
        @type exchange_name: str

        @return: List of all currency-pairs for the given exchange.
        @rtype: list[ExchangeCurrencyPair]
        """
        with self.session_scope() as session:
            currency_pairs = list()
            exchange_id: int = session.query(Exchange.id).filter(Exchange.name == exchange_name.upper()).scalar()
            if exchange_id is not None:
                currency_pairs = session.query(ExchangeCurrencyPair).filter(
                    ExchangeCurrencyPair.exchange_id == exchange_id).all()
                session.expunge_all()

        return currency_pairs

    def get_currency_pairs_with_first_currency(self, exchange_name: str, currency_names: List[str]) \
            -> List[ExchangeCurrencyPair]:
        """
        Returns all currency-pairs for the given exchange that have any of the given currencies
        as the first currency.

        @param exchange_name: Name of the exchange.
        @type exchange_name: str
        @param currency_names: List of the currency names that are viable as first-currencies.
        @type currency_names: list[str]

        @return: List of the currency-pairs which start with any of the currencies in currency_names
                 on the given exchange.
                 List is empty if there are no currency pairs in the database which fulfill the requirements.
        @rtype: list[ExchangeCurrencyPair]
        """
        if isinstance(currency_names, str):
            currency_names = [currency_names]
        all_found_currency_pairs: List[ExchangeCurrencyPair] = list()
        if exchange_name is not None and exchange_name:
            exchange_id: int = self.get_exchange_id(exchange_name)

            with self.session_scope() as session:
                if currency_names is not None:
                    for currency_name in currency_names:
                        if currency_name is not None and currency_name:
                            first_id: int = self.get_currency_id(currency_name)

                            found_currency_pairs = session.query(ExchangeCurrencyPair).filter(
                                ExchangeCurrencyPair.exchange_id == exchange_id,
                                ExchangeCurrencyPair.first_id == first_id).all()

                            if found_currency_pairs is not None:
                                all_found_currency_pairs.extend(found_currency_pairs)
                session.expunge_all()
        return all_found_currency_pairs

    def get_currency_pairs_with_second_currency(self, exchange_name: str, currency_names: List[str]) \
            -> List[ExchangeCurrencyPair]:
        """
        Returns all currency-pairs for the given exchange that have any of the given currencies
        as the second currency.

        @param exchange_name: Name of the exchange.
        @type exchange_name: str
        @param currency_names: List of the currency names that are viable as second currencies.
        @type currency_names: list[str]

        @return: List of the currency-pairs which end with any of the currencies in currency_names
                 on the given exchange.
                 List is empty if there are no currency pairs in the database which fulfill the requirements.
        @rtype: list[ExchangeCurrencyPair]
        """
        all_found_currency_pairs: List[ExchangeCurrencyPair] = list()
        if exchange_name:
            exchange_id: int = self.get_exchange_id(exchange_name)

            with self.session_scope() as session:
                if currency_names is not None:
                    for currency_name in currency_names:
                        if currency_name:
                            second_id: int = self.get_currency_id(currency_name)

                            found_currency_pairs = session.query(ExchangeCurrencyPair).filter(
                                ExchangeCurrencyPair.exchange_id == exchange_id,
                                ExchangeCurrencyPair.second_id == second_id).all()

                            if found_currency_pairs is not None:
                                all_found_currency_pairs.extend(found_currency_pairs)

                session.expunge_all()
        return all_found_currency_pairs

    def get_readable_query(self,
                           db_table: DatabaseTable,
                           query_everything: bool,
                           from_timestamp: datetime = None,
                           to_timestamp: datetime = TimeHelper.now(),
                           exchanges: List[str] = None,
                           currency_pairs: List[str] = None,
                           first_currencies: List[str] = None,
                           second_currencies: List[str] = None) -> DataFrame:

        """ #ToDo Debug, not working when filters are applied.
         Queries based on the parameters readable database data and returns it.
         If query_everything is true, everything ticker tuple will be returned.
         This is also the case if query_everything is false but there were no
         exchanges or currencies/currency pairs given.
         If exchanges are given only tuples of these given exchanges will be returned.
         If there are no currencies/currency pairs given,
         all ticker-tuple of the given exchange will be returned.
         If currencies are given note that only ticker tuple with currency pairs,
         which have either any currency in first_currencies as first OR any currency
         in second_currencies as second OR any currency pairs in currency_pairs will be returned.
         If timestamps are given the queried tuples will be filtered accordingly.

         So query logic for each tuple is (if exchange, currencies and time are given):
             exchange AND (first OR second OR pair) AND from_time AND to_time

         See csv-config for details of how to write/give parameters.

         @param db_table: The respective object of the table to be queried.
         @type db_table: Union[HistoricRate, OrderBook, Ticker, Trade]
         @param query_everything: If everything in the database should be queried.
         @type query_everything: bool
         @param from_timestamp: Minimum date for the start of the request.
         @type from_timestamp: datetime
         @param to_timestamp: Maximum date for the start of the request.
         @type to_timestamp: datetime
         @param exchanges: List of exchanges of which the tuple should be queried.
         @type exchanges: list[str]
         @param currency_pairs: List of specific currency pairs that should be queried.
                                Dict needs to have the following structure:
                                   - first: 'Name of the first currency'
                                   - second: 'Name of the second currency'
         @type currency_pairs: list[dict[str, str]]
         @param first_currencies: List of viable currencies for the first currency in a currency pair.
         @type first_currencies: list[str]
         @param second_currencies: List of viable currencies for the second currency in a currency pair.
         @type second_currencies: list[str]

         @return: DataFrame of readable database tuple.
                  DataFrame might be empty if database is empty or there where no ExchangeCurrencyPairs
                  which fulfill the above stated requirements.
         @rtype: Pandas DataFrame
         """
        with self.session_scope() as session:
            first = aliased(Currency)
            second = aliased(Currency)

            data: Query = session.query(Exchange.name.label("exchange"),
                                        first.name.label("first_currency"),
                                        second.name.label("second_currency"),
                                        db_table). \
                join(ExchangeCurrencyPair, db_table.exchange_pair_id == ExchangeCurrencyPair.id). \
                join(Exchange, ExchangeCurrencyPair.exchange_id == Exchange.id). \
                join(first, ExchangeCurrencyPair.first_id == first.id). \
                join(second, ExchangeCurrencyPair.second_id == second.id)

            if query_everything:
                result = pd_read_sql_query(data.statement, con=session.bind)
                session.expunge_all()
                return result

            # ensure parameters are in a list
            if not all([isinstance(item, list) for item in [exchanges, first_currencies, second_currencies]]):
                if exchanges and isinstance(exchanges, str):
                    exchanges = split_str_to_list(exchanges)
                if first_currencies and isinstance(first_currencies, str):
                    first_currencies = split_str_to_list(first_currencies)
                if second_currencies and isinstance(second_currencies, str):
                    second_currencies = split_str_to_list(second_currencies)
                if currency_pairs and isinstance(currency_pairs, str):
                    currency_pairs = split_str_to_list(currency_pairs)

            if exchanges:
                exchange_names = [name.upper() for name in exchanges]
                data = data.filter(Exchange.name.in_(exchange_names))

            if first_currencies:
                first_currency_names = [name.upper() for name in first_currencies]
                data = data.filter(first.name.in_(first_currency_names))
            if second_currencies:
                second_currency_names = [name.upper() for name in second_currencies]
                data = data.filter(second.name.in_(second_currency_names))
            if currency_pairs:
                currency_pairs_names = [(pair.split("-")[0].upper(), pair.split("-")[1].upper()) for pair in
                                        currency_pairs]
                currency_pairs_names = [[item.replace(" ", "") for item in pair] for pair in currency_pairs_names]

                data = data.filter(tuple_(first.name, second.name).in_(currency_pairs_names))

            if from_timestamp:
                data = data.filter(db_table.time >= from_timestamp)
            if to_timestamp:
                data = data.filter(db_table.time <= to_timestamp)

            # TODO: Philipp: Ask Steffen because result is not used?

            result = pd_read_sql_query(data.statement, con=session.bind)
        session.expunge_all()
        return result

    def get_or_create_exchange_pair_id(self,
                                       exchange_name: str,
                                       first_currency_name: str,
                                       second_currency_name: str,
                                       is_exchange: bool) -> int:
        """
        Returns an existing exchange-currency-pair id or creates a new instance and returns the id.

        @param exchange_name: Exchange name
        @param first_currency_name: First currency name
        @param second_currency_name: Second currency name
        @param is_exchange: Is from exchange or platform

        @return: Id of the existing or newly created exchange currency pair
        """
        temp_currency_pair = {"exchange_name": exchange_name,
                              "first_currency_name": first_currency_name,
                              "second_currency_name": second_currency_name}

        with self.session_scope() as session:
            currency_pair: ExchangeCurrencyPair = self._get_exchange_currency_pair(session, **temp_currency_pair)

            if not currency_pair and all([exchange_name, first_currency_name, second_currency_name, is_exchange]):
                self._persist_exchange_currency_pair(is_exchange=is_exchange, **temp_currency_pair)
                return self.get_or_create_exchange_pair_id(is_exchange=is_exchange, **temp_currency_pair)
            else:
                return currency_pair.id

    def get_first_timestamp(self, table: DatabaseTable, exchange_pair_id: int, last_row_id: int) -> datetime:
        """
        Returns the earliest timestamp from the specified table if the latest timestamp is less than 2 days old.
        If the table is empty, the method trys to catch information from the helper table PairInfo.
        Otherwise the timestamp from now.

        @param table: The database table to be queried.
        @type table: Union[HistoricRate, OrderBook, Ticker, Trade]
        @param exchange_pair_id: The exchange_pair_id of interest.
        @type exchange_pair_id: int
        @param last_row_id: The row-id of the last entry of previous request.
        @type: last_row_id: int
        @return: datetime: Earliest timestamp of specified table or timestamp from now.
        @rtype: datetime
        """
        with self.session_scope() as session:
            if last_row_id:
                timestamp = session.execute(f"SELECT time FROM historic_rates where rowid = {last_row_id} "
                                            f"ORDER BY time DESC").first()[0]
                return TimeHelper.from_timestamp(timestamp, TimeUnit.MILLISECONDS)

            (earliest_timestamp,) = session \
                .query(func.min(table.time)) \
                .filter(table.exchange_pair_id == exchange_pair_id) \
                .first()
            (newest_timestamp,) = session \
                .query(func.max(table.time)) \
                .filter(table.exchange_pair_id == exchange_pair_id) \
                .first()

        # two days as some exchanges lag behind one day for historic_rates
        if earliest_timestamp and (TimeHelper.now() - newest_timestamp) < timedelta(days=2):
            return earliest_timestamp

        return TimeHelper.now()

    def persist_exchange(self, exchange_name: str, is_exchange: bool) -> None:
        """
        Persists the given exchange-name if it's not already in the database.

        @param exchange_name:
            Name that should is to persist.
        @param is_exchange: boolean indicating if the exchange is indeed an exchange or a platform
        """
        with self.session_scope() as session:
            exchange_id = session.query(Exchange.id).filter(Exchange.name == exchange_name.upper()).first()
            if exchange_id is None:
                exchange = Exchange(name=exchange_name, is_exchange=is_exchange)
                session.add(exchange)

    def _persist_exchange_currency_pair(self,
                                        exchange_name: str,
                                        first_currency_name: str,
                                        second_currency_name: str,
                                        is_exchange: bool) -> None:
        """
        Adds a single ExchangeCurrencyPair to the database is it does not already exist.

        @param exchange_name: str
            Name of the exchange.
        @param first_currency_name: str
            Name of the first currency.
        @param second_currency_name: str
            Name of the second currency.
        @param is_exchange: boolean indicating if the exchange is indeed an exchange or a platform
        """
        self.persist_exchange_currency_pairs([(exchange_name, first_currency_name, second_currency_name)],
                                             is_exchange=is_exchange)

    def persist_exchange_currency_pairs(self, currency_pairs: Iterable[Tuple[str, str, str]],
                                        is_exchange: bool) -> None:
        """
        Persists the given already formatted ExchangeCurrencyPair-tuple if they not already exist.
        The formatting ist done in @see{Exchange.format_currency_pairs()}.

        Tuple needs to have the following structure:
            (exchange-name, first currency-name, second currency-name)

        @param currency_pairs:
            Iterator of currency-pair tuple that are to persist.
        @param is_exchange: boolean indicating if the exchange is indeed an exchange or a platform
        """
        if currency_pairs is not None:

            i = 0
            with self.session_scope() as session:
                for currency_pair in currency_pairs:
                    exchange_name = currency_pair[0]
                    first_currency_name = currency_pair[1]
                    second_currency_name = currency_pair[2]
                    i += 1

                    if any([exchange_name, first_currency_name, second_currency_name]) is None:
                        continue

                    if first_currency_name == second_currency_name:
                        continue

                    existing_exchange = session.query(Exchange).filter(Exchange.name == exchange_name.upper()).first()
                    exchange: Exchange = existing_exchange if existing_exchange is not None else Exchange(
                        name=exchange_name, is_exchange=is_exchange)

                    existing_first_cp = session.query(Currency).filter(
                        Currency.name == first_currency_name.upper()).first()
                    first: Currency = existing_first_cp if existing_first_cp is not None else Currency(
                        name=first_currency_name, from_exchange=is_exchange)

                    existing_second_cp = session.query(Currency).filter(
                        Currency.name == second_currency_name.upper()).first()
                    second: Currency = existing_second_cp if existing_second_cp is not None else Currency(
                        name=second_currency_name, from_exchange=is_exchange)

                    existing_exchange_pair = session.query(ExchangeCurrencyPair).filter(
                        ExchangeCurrencyPair.exchange_id == exchange.id,
                        ExchangeCurrencyPair.first_id == first.id,
                        ExchangeCurrencyPair.second_id == second.id)

                    existing_exchange_pair = session.query(existing_exchange_pair.exists()).scalar()

                    if not existing_exchange_pair:
                        exchange_pair = ExchangeCurrencyPair(exchange=exchange, first=first, second=second)
                        session.add(exchange_pair)
                    # persist data every 500 CPs in order to avoid slowing down
                    if i % 500 == 0:
                        session.commit()

    def persist_response(self,
                         exchanges_with_pairs: Dict[Exchange, Dict[ExchangeCurrencyPair, Optional[int]]],
                         exchange: Exchange,
                         db_table: DatabaseTable,
                         formatted_response: Iterator[Any],
                         update_on_conflict: bool = False) -> Dict[ExchangeCurrencyPair, Optional[int]]:
        """
        Method to persist the formatted response into the database. Every data tuple get is inspected for
        valid data, i.e. the mapping key and corresponding data must be one of the table columns of the database.
        Furthermore, if the data tuple does not contain an exchange_pair_id, a new ExchangeCurrencyPair is
        persisted into the database. To persist data, this method makes use of:
            - sqlalchemy.dialects.<dialect>.insert
        which specifically allows for:
            - on_conflict_do_nothing, or
            - on_conflict_do_update.
        By default, conflicts are ignored, however, if existing rows are supposed to be updated,
        the behaviour can be changed.

        @param exchanges_with_pairs: Dict containing all requested exchanges and pairs
        @param exchange: Exchange Object
        @param db_table: Affected database table, i.e. request-method
        @param formatted_response: Generator of extracted and formatted response.
        @param update_on_conflict: Bool if conflicting rows should be updated or ignored
        @return: Dict containing ExchangeCurrencyPair-Object and the last inserted row_id
        """
        col_names = [key.name for key in inspect(db_table).columns]
        primary_keys = [key.name for key in inspect(db_table).primary_key]
        counter_dict: Dict[int, int] = dict()
        requested_cp_ids = [pair.id for pair in exchanges_with_pairs[exchange]]

        while True:
            try:
                data_to_persist: List[dict, ...] = list()
                data, mappings = next(formatted_response)

                for data_tuple in data:
                    data_tuple = dict(zip(mappings, data_tuple))

                    if "exchange_pair_id" not in data_tuple.keys():
                        temp_pair = {"exchange_name": exchange.name,
                                     "first_currency_name": data_tuple["currency_pair_first"],
                                     "second_currency_name": data_tuple["currency_pair_second"],
                                     "is_exchange": exchange.is_exchange}

                        new_pair_id = self.get_or_create_exchange_pair_id(**temp_pair)
                        if new_pair_id in requested_cp_ids:
                            data_tuple.update({"exchange_pair_id": new_pair_id})
                        else:
                            continue

                    data_tuple = {key: data_tuple.get(key, None) for key in col_names}
                    data_to_persist.append(data_tuple)
                    exchange_pair_id = [item.get("exchange_pair_id") for item in data_to_persist]
                    # remove duplicates
                    exchange_pair_id = list(dict.fromkeys(exchange_pair_id))

                if not data_to_persist:
                    continue

                # Sort data by timestamp in order to ensure the last_row_id (see below) to be with the oldest timestamp.
                # This is used for historic_rates.get_first_timestamp(), if the oldest timestamp of the previous
                # request is wanted, instead of the oldest timestamp in the database.
                data_to_persist = sorted(data_to_persist, key=lambda i: i["time"], reverse=True)

                with self.session_scope() as session:

                    stmt = self.insert_module.insert(db_table).values(data_to_persist)

                    try:
                        stmt = stmt.on_conflict_do_nothing(index_elements=primary_keys)
                    except AttributeError:
                        stmt = stmt.prefix_with("IGNORE")

                    row_count = session.execute(stmt)

                    print(f"Pair-ID {exchange_pair_id[0] if len(exchange_pair_id) == 1 else 'ALL'}"
                          f" - {exchange.name.capitalize()}: {row_count.rowcount} tuple(s)")

                    # Dict containing the ExchangeCurrencyPair as key and the last_row_id as value, if and only if
                    # at least self._min_return_tuples are persisted. If not, the ExchangeCurrencyPair will be kicked
                    # out in the next run. The strange subscription of the dict-comprehension is because of the nested
                    # dict in exchange_with_pairs: Dict[Exchange, Dict[ExchangeCurrencyPair, Optional[int]]].
                    counter_dict.update({k: row_count.lastrowid for k, v in exchanges_with_pairs[exchange].items()
                                         if k.id in exchange_pair_id and row_count.rowcount >= self._min_return_tuples})

            except StopIteration:
                break

        return counter_dict if counter_dict else {}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module providing additional exception classes.

Classes:
    - MappingNotFoundException
    - DifferentExchangeContentException
    - NoCurrencyPairProvidedException
"""


class MappingNotFoundException(Exception):
    """
    Custom exception that is thrown when needed mappings could not be found. Most likely this happens when there is a
    typo in the exchange-config file or the file is incomplete.
    """

    def __init__(self, exchange_name: str, mapping_name: str):
        """
        @param exchange_name: Current exchange the exception is thrown for.
        @type exchange_name: str
        @param mapping_name: Name of the method/mappings that are needed.
        @type mapping_name: str
        """
        super().__init__(f"No mapping with the name '{mapping_name}' for the exchange '{exchange_name}' found.")


class DifferentExchangeContentException(Exception):
    """Custom exception that is thrown when the given content is from a different exchange."""

    def __init__(self, name_of_given_exchange: str, name_of_receiving_exchange: str):
        """
        @param name_of_given_exchange: Name of the exchange that the content is from.
        @type name_of_given_exchange: str
        @param name_of_receiving_exchange: Nome of the exchange that the content is handed to.
        @type name_of_receiving_exchange: str
        """
        super().__init__(f"'{name_of_receiving_exchange}' was given content from '{name_of_given_exchange}'.")


class NoCurrencyPairProvidedException(Exception):
    """
    Custom exception that is thrown when there was no currency_pair_first or _second found or provided in the
    exchange-config but is needed.
    """

    def __init__(self, exchange_name: str, method_name: str):
        """
        @param exchange_name: Name of the exchange that had no currency pair provided.
        @type exchange_name: str
        @param method_name: Name of the method where the currency pair is needed.
        @type method_name: str
        """
        super().__init__(f"For the exchange '{exchange_name}' for the method '{method_name}' currency pair info is"
                         f"needed but not sufficiently provided.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module is a wrapper around the whole package. Its main function is to export all relevant configuration files
to the current working directory of the user, start the program, establish database connections and export data
into csv-files.
"""
import os
import shutil
from pathlib import Path
from typing import Any, Optional, Dict
import pandas as pd
from sqlalchemy.orm.session import Session

# noinspection PyUnresolvedReferences
try:
    from open_crypto import _paths  # pylint: disable=unused-import
except (ImportError, Exception):
    import _paths  # pylint: disable=unused-import
import main
from model.utilities.kill_switch import KillSwitch
from model.utilities.export import CsvExport, database_session
from model.utilities.utilities import read_config, get_all_exchanges_and_methods, prepend_spaces_to_columns
from model.utilities.settings import Settings  # pylint: disable=unused-import
from model.utilities.github_downloader import GitDownloader
from model.database.tables import *  # pylint: disable=unused-import
from examples import Examples  # pylint: disable=unused-import


def update_maps() -> None:
    """
    Downloads the most recent exchange mappings from Github and saves them within the package directory.
    If the CWD differs from the package directory (i.e. the site-packages directory), further copies the updated
    resources to the current working directory.
    """

    GitDownloader.main()
    if os.getcwd() != _paths.all_paths.get("package_path"):
        copy_resources()


def _check_path(path: Union[Path, str], check_only: bool = False) -> bool:
    """
    Checks if all resources are in the current working directory. If not, calls the function update_maps()

    @param path: The path.
    @param check_only: Only check if path exists without copying resources.
    @type path: str
    @return True if path exists, else either return False or copy_resources.
    """
    destination = Path(path).joinpath("resources")
    if os.path.exists(destination):
        return True
    if check_only:
        return False
    copy_resources()


def copy_resources(directory: Union[Path, str] = os.getcwd()) -> None:
    """
    Copies everything from the folder "resources" into the current working directory. If files already exist,
    the method will override them (i.e. first delete and then copy).

    @param directory: The directory.
    @type directory: Current working directory
    """

    # source = os.path.dirname(os.path.realpath(__file__)) + "/resources"
    source = Path(_paths.all_paths.get("package_path")).joinpath("/resources").__str__()
    destination = Path(directory).joinpath("resources").__str__()

    for src_dir, dirs, files in os.walk(source):
        dst_dir = src_dir.replace(source, destination, 1)
        try:
            dirs.remove("templates")
            dirs.remove("__pycache__")
        except ValueError:
            pass

        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for file_ in files:
            src_file = os.path.join(src_dir, file_)
            dst_file = os.path.join(dst_dir, file_)
            if os.path.exists(dst_file):
                # in case of the src and dst are the same file
                if os.path.samefile(src_file, dst_file):
                    continue
                os.remove(dst_file)
            if not src_file.endswith(".py"):
                shutil.copy(src_file, dst_dir)
    print(f"\nCopied resources to '{directory}'.")


def get_session(filename: str, db_path: str = os.getcwd()) -> Session:
    """
    Returns an open SqlAlchemy-Session. The session is obtained from the DatabaseHandler via the module export.py.
    Furthermore this functions imports all database defining classes to work with.

    @param filename: Name of the configuration file to init the DatabaseHandler
    @type filename: str
    @param db_path: path to the database. Default: current working directory
    @type db_path: str

    @return: A database session.
    @rtype: Session
    """
    return database_session(filename=filename, db_path=db_path)


def exchanges_and_methods(return_dataframe: bool = False) -> Optional[pd.DataFrame]:
    """
    Lists all exchanges and methods.

    @param return_dataframe: If True, the dataframe is returned.
    @type return_dataframe: boolean

    @return: Print or return dataframe
    @rtype: Optional[pd.DataFrame]
    """
    working_directory = Path(os.getcwd())
    _check_path(working_directory)

    dataframe = pd.DataFrame.from_dict(get_all_exchanges_and_methods())
    dataframe.index.name = 'Exchange'
    pd.set_option("display.max_rows", 500)

    if return_dataframe:
        return dataframe.transpose()
    else:
        return prepend_spaces_to_columns(dataframe.transpose(), 3)


def get_config(filename: Optional[str] = None) -> Dict[str, Any]:
    """
    Parses the specified configuration file.

    @param filename: The name of the configuration file.
    @type filename: Optional[str]

    @return: Returns the parsed configuration.
    @rtype: dict[str, Any]
    """
    return read_config(file=filename)


def get_config_template(csv: bool = False) -> None:
    """
    Creates a copy of the config templates and puts it into the resources/configs folder.

    @param csv: If True, create an csv-export config. Else create a config for the runner.
    @type csv: bool
    """
    if csv:
        filename = "csv_export_template.yaml"
    else:
        filename = "request_template.yaml"

    path_exists = os.path.exists(_paths.all_paths.get("user_config_path"))
    if not path_exists:
        print("Copy all resources to your current working directory first.\n"
              "Call 'runner.update_maps()' for this task.")
        return

    source = _paths.all_paths.get("template_path")
    destination = _paths.all_paths.get("user_config_path")

    # Uncomment if the above paths show any malfunction.
    # source = os.path.dirname(os.path.realpath(__file__)) + "/resources/templates"
    # destination = os.getcwd() + "/resources/configs/user_configs"

    if os.path.exists(destination.joinpath(filename)):
        os.remove(destination.joinpath(filename))

    shutil.copy(source.joinpath(filename),
                destination.joinpath(filename))
    print(f"Created new config template. \nLocation: '{os.path.realpath(destination)}'.")


def export(file: Optional[str] = None, file_format: str = "csv", *args: Any, **kwargs: Any) -> None:
    """
    Calls the imported module CsvExport and the respective method create_csv(). This will take a csv-export config as
    input and write data into a csv-file depending on the configuration.

    @param file: The name of the export file.
    @type file: Optional[str]
    @param file_format: The file format of the export file.
    @type file_format: str
    """
    CsvExport(file).export(data_type=file_format, *args, **kwargs)


def run(configuration_file: Optional[str] = None, kill_after: int = None) -> None:
    """
    Starts the program after checking if all necessary folder are available (i.e. config and yaml-maps).

    @param configuration_file: The configuration file.
    @type configuration_file: Optional[str]
    @param kill_after: Kills the thread.
    @type kill_after: int
    """
    working_directory = os.getcwd()
    _check_path(Path(working_directory))

    if kill_after and isinstance(kill_after, int):
        switch: KillSwitch
        with KillSwitch() as switch:
            switch.set_timer(kill_after)
    try:
        main.run(configuration_file, working_directory)
    except SystemExit:
        pass


if __name__ == '__main__':
    run('request_template')


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module provides the database defining classes as well as specific views on all unreadable database tables.

Classes:
 - Exchange: Defines the exchanges table.
 - Currency: Defines the currencies table.
 - ExchangesCurrencyPair: Defines the exchanges_currency_pairs table.
 - Ticker: Defines the tickers tables.
 - Trade: Defines the trades table.
 - HistoricRate: Defines the historic_rates table.
 - OrderBook: Defines the order_books table.

 - ExchangeCurrencyPairView: Provides a view on the exchanges_currency_pairs table.
 - TickerView: Provides a view on the tickers table.
 - TradeView: Provides a view on the trades table.
 - HistoricRateView: Provides a view on the historic_rates table.
 - OrderBookView: Provides a view on the order_books table.
"""

from typing import Union, Type

from sqlalchemy import Column, Integer, String, Boolean, ForeignKey, CheckConstraint, Float, select, BigInteger
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, validates, aliased
from sqlalchemy_utils import create_view

from model.database.type_decorators import UnixTimestampMs

Base = declarative_base()  # pylint: disable=invalid-name
metadata = Base.metadata


class Exchange(Base):
    """
    Database ORM-Class storing the exchange table. ALl exchange used to perform requests
    are listed in this table.

    id: int
        Auto incremented unique identifier.
    name: str
        The explicit name of the exchange defined in the .yaml-file.
    """
    __tablename__ = "exchanges"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name: Column = Column(String(50), nullable=False, unique=True)
    active = Column(Boolean, default=True)
    is_exchange = Column(Boolean, default=True)
    exceptions = Column(Integer, unique=False, nullable=True, default=0)
    total_exceptions = Column(Integer, unique=False, nullable=True, default=0)

    def __repr__(self) -> str:
        return f"#{self.id}: {self.name}, Active: {self.active}"

    @validates("name")
    def convert_upper(self, _: str, value: str) -> str:
        """
        Converts strings into upper cases.
        """
        return value.upper()


class Currency(Base):
    """
    Database ORM-Class storing all currencies.

    id: int
        Auto incremented unique identifier.
    name: str
        Name of the currency written out.
    symbol: str
        Abbreviation of the currency
    """
    __tablename__ = "currencies"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), unique=True, nullable=False)
    from_exchange = Column(Boolean, default=True)

    def __repr__(self) -> str:
        return f"#{self.id}: {self.name}"

    @validates("name")
    def convert_upper(self, _: str, value: str) -> str:
        """
        Converts strings into upper cases.
        """
        return value.upper()


class ExchangeCurrencyPair(Base):
    """
    Database ORM-Class storing the ExchangeCurrencyPair.

    exchange_id: int
        The unique id of each exchange taken from the ForeignKey.
    first_id: int
        The unique id of each currency_pair taken from the table Currency.
    second_id: int
        The unique id of each currency_pair taken from the table Currency.

    exchange: relationship
        Relationship with table Exchange
    first: relationship
        Relationship with table Currency
    second: relationship
        Relationship with table Currency
    __table_args__:
        First ID must be unequal to Second ID.
    """
    __tablename__ = "exchanges_currency_pairs"

    id = Column(Integer, primary_key=True, autoincrement=True)
    exchange_id = Column(Integer, ForeignKey("exchanges.id"))
    first_id = Column(Integer, ForeignKey("currencies.id"))
    second_id = Column(Integer, ForeignKey("currencies.id"))

    exchange = relationship("Exchange", backref="exchanges_currency_pairs", lazy="joined")
    first = relationship("Currency", foreign_keys="ExchangeCurrencyPair.first_id", lazy="joined")
    second = relationship("Currency", foreign_keys="ExchangeCurrencyPair.second_id", lazy="joined")

    __table_args__ = (CheckConstraint(first_id != second_id),)

    def __repr__(self) -> str:
        return f"#{self.id}: {self.exchange.name}({self.exchange_id}), " \
               f"{self.first.name}({self.first_id})-{self.second.name}({self.second_id})"

    def __str__(self) -> str:
        return self.__repr__()


class Ticker(Base):
    """
    Database ORM-Class storing the ticker data.

    exchange_pair_id: int
        Unique Exchange_Currency_Pair identifier. The exchange_pair_id is used by the database_handler to check
         for existing exchange_currency_pairs. If not existing, the currency and/or exchange is created.

    exchange_pair: relationship
        The corresponding relationship table with ExchangeCurrencyPair

    start_time: UnixTimestamp
        Timestamp of the execution of an exchange request (UTC). Timestamps are unique for each exchange.

    response_time: UnixTimestamp
        Timestamp of the response. Timestamps are created by the OS, the delivered ones from the exchange are not used.
        Timestamps are equal for each exchange for one run (resulting in an error of approx. 5 seconds average)
         to ease data usage later.
        Timestamps are rounded to seconds (UTC)

    last_price: float
        Latest price of the currency_pair given from the exchange.
    best_ask: float
        Best ask price of an exchange for a currency_pair.
    best_bid: float
        Best bid price of an exchange for a currency_pair.
    daily_volume: float
        The traded volume of an currency_pair on an exchange. Definition can differ for each exchange!

    """
    __tablename__ = "tickers"

    exchange_pair_id = Column(Integer, ForeignKey("exchanges_currency_pairs.id"), primary_key=True)
    exchange_pair = relationship("ExchangeCurrencyPair", backref="tickers")

    start_time = Column(UnixTimestampMs)
    time = Column(UnixTimestampMs, primary_key=True)
    last_price = Column(Float)
    best_ask = Column(Float)
    best_bid = Column(Float)

    def __repr__(self) -> str:
        return f"#{self.exchange_pair_id}, {self.exchange_pair.exchange.name}: " \
               f"{self.exchange_pair.first.name}-{self.exchange_pair.second.name}, ${self.last_price} at {self.time}"


class HistoricRate(Base):
    """
    Table for the method historic_rates. Tables contains the exchange_currency_pair_id, gathered from the
    foreign_keys.

    Primary_keys are Exchange_Pair_id and the timestamp.

    Table contains standard OHLCV values (Open - High - Low - Close - Volume24h)

    __repr__(self) describes the representations of the table if queried. The database will return the
    object as normal, but print "ID, Exchange: First-Second, $ Close at time" in clear names for better
    readability.
    """
    __tablename__ = "historic_rates"

    exchange_pair_id = Column(Integer, ForeignKey("exchanges_currency_pairs.id"), primary_key=True)
    exchange_pair = relationship("ExchangeCurrencyPair", backref="historic_rates")
    time = Column(UnixTimestampMs, primary_key=True)
    open = Column(Float)
    high = Column(Float)
    low = Column(Float)
    close = Column(Float)
    volume = Column(Float)
    # base_volume = Column(Float)  # ToDo: check if base_volume is correct.
    market_cap = Column(Float)

    def __repr__(self) -> str:
        return f"ID {self.exchange_pair_id}, {self.exchange_pair.exchange.name}: " \
               f"{self.exchange_pair.first.name}-{self.exchange_pair.second.name}, close {self.close} at {self.time}"


class Trade(Base):
    """
    Table for the method trades. Tables contains the exchange_currency_pair_id, gathered from the
    foreign_keys.
    Primary_keys are Exchange_Pair_id and the timestamp.

    Table contains the last trades, trade amount, trade direction (buy/sell) and timestamp.

    __repr__(self) describes the representation if queried.

    """
    __tablename__ = "trades"

    exchange_pair_id = Column(Integer, ForeignKey("exchanges_currency_pairs.id"), primary_key=True)
    exchange_pair = relationship("ExchangeCurrencyPair", backref="trades")
    id = Column(BigInteger, primary_key=True)
    time = Column(UnixTimestampMs, primary_key=True)

    amount = Column(Float, primary_key=True)
    best_bid = Column(Float)
    best_ask = Column(Float)
    price = Column(Float)
    _direction = Column("direction", String(5)) # This is changes from Integer to String as the Hybrid Property method
                                                # does currently not work with insert().on_conflict_do_nothing().
                                                # It Does however work with the standard session.add(Some_Instance)

    @hybrid_property
    def direction(self) -> Union[str, int]:
        """
        Returns the direction.
        #ToDo: Make Hybrid Property work again with insert.on_conflict_do_nothing() method.
        """
        return self._direction

    @direction.setter
    def direction(self, direction: Union[str, int]) -> None:
        """
        Converts the string representation of the trade direction, i.e. 'sell' or 'buy', into an integer.

        @param: String representation of the direction.
        """
        if isinstance(direction, str):
            if direction.lower() == "sell":
                self._direction = 0
            elif direction.lower() == "buy":
                self._direction = 1
            else:
                self._direction = direction

    def __repr__(self) -> str:
        return f"Last Transaction: {self.exchange_pair.exchange.name}, {self.exchange_pair.first.name}-" \
               f"{self.exchange_pair.second.name}: {self.amount} for {self.price} at {self.time}"

    @validates("direction")
    def convert_upper(self, _: str, value: str) -> str:
        """
        Converts strings into upper cases.
        """
        return value.upper()


class OrderBook(Base):
    """
    Table for the method order-books. Tables contains the exchange_currency_pair_id, gathered from the
    foreign_keys.

    Primary_keys are Exchange_Pair_id, id, and position.

    Table next to the bids and asks (both with Price and Amount) the position which indicates the position in
    the order book at given time. I.e position 0 contains the highest Bid and the lowest Ask. The ID is gathered
    directly from the exchange and is used to identify to identify changes in the order-book.
    """
    __tablename__ = "order_books"

    exchange_pair_id = Column(Integer, ForeignKey("exchanges_currency_pairs.id"), primary_key=True)
    exchange_pair = relationship("ExchangeCurrencyPair", backref="OrderBook")

    id = Column(BigInteger, primary_key=True)
    position = Column(Integer, primary_key=True)

    time = Column(UnixTimestampMs)
    bids_amount = Column(Float)
    bids_price = Column(Float)
    asks_price = Column(Float)
    asks_amount = Column(Float)


class ExchangeCurrencyPairView(Base):
    """
    View vor ExchangeCurrencyPairs.
    """
    first = aliased(Currency)
    second = aliased(Currency)

    __table__ = create_view(
        name="exchanges_currency_pairs_view",
        selectable=select(
            [
                ExchangeCurrencyPair.id,
                Exchange.name.label("exchange_name"),
                first.name.label("first_name"),
                second.name.label("second_name"),
            ],
            from_obj=(
                ExchangeCurrencyPair.__table__.join(Exchange, ExchangeCurrencyPair.exchange_id == Exchange.id)
                                              .join(first, ExchangeCurrencyPair.first_id == first.id)
                                              .join(second, ExchangeCurrencyPair.second_id == second.id)
            )
        ),
        metadata=Base.metadata
    )


class TickerView(Base):
    """
    View for Tickers.
    Instead of only showing the ID of the ExchangeCurrencyPair the View displays
    the exchange name and the name of the first and second currency.
    """
    first = aliased(Currency)
    second = aliased(Currency)
    __table__ = create_view(
        name="tickers_view",
        selectable=select(
            [
                Exchange.name.label("exchange"),
                first.name.label("first_currency"),
                second.name.label("second_currency"),
                Ticker.start_time,
                Ticker.time,
                Ticker.last_price,
                Ticker.best_ask,
                Ticker.best_bid,
                # Ticker.daily_volume
            ],
            from_obj=(
                Ticker.__table__.join(ExchangeCurrencyPair, Ticker.exchange_pair_id == ExchangeCurrencyPair.id)
                                .join(Exchange, ExchangeCurrencyPair.exchange_id == Exchange.id)
                                .join(first, ExchangeCurrencyPair.first_id == first.id)
                                .join(second, ExchangeCurrencyPair.second_id == second.id)
            )
        ),
        metadata=Base.metadata
    )


class TradeView(Base):
    """
    View for Trades.
    Instead of only showing the ID of the ExchangeCurrencyPair the View displays
    the exchange name and the name of the first and second currency.
    """
    first = aliased(Currency)
    second = aliased(Currency)

    __table__ = create_view(
        name="trades_view",
        selectable=select(
            [
                Exchange.name.label("exchange"),
                first.name.label("first_currency"),
                second.name.label("second_currency"),
                Trade.id,
                Trade.time,
                Trade.amount,
                Trade.best_ask,
                Trade.best_bid,
                Trade.price,
                Trade.direction,
            ],
            from_obj=(
                Trade.__table__.join(ExchangeCurrencyPair, Trade.exchange_pair_id == ExchangeCurrencyPair.id)
                               .join(Exchange, ExchangeCurrencyPair.exchange_id == Exchange.id)
                               .join(first, ExchangeCurrencyPair.first_id == first.id)
                               .join(second, ExchangeCurrencyPair.second_id == second.id)
            )
        ),
        metadata=Base.metadata
    )


class OrderBookView(Base):
    """
    View for Order-Books.
    Instead of only showing the ID of the ExchangeCurrencyPair the View displays
    the exchange name and the name of the first and second currency.
    """
    first = aliased(Currency)
    second = aliased(Currency)
    __table__ = create_view(
        name="order_books_view",
        selectable=select(
            [
                Exchange.name.label("exchange"),
                first.name.label("first_currency"),
                second.name.label("second_currency"),
                OrderBook.id,
                OrderBook.position,
                OrderBook.time,
                OrderBook.bids_amount,
                OrderBook.bids_price,
                OrderBook.asks_price,
                OrderBook.asks_amount
            ],
            from_obj=(
                OrderBook.__table__.join(ExchangeCurrencyPair, OrderBook.exchange_pair_id == ExchangeCurrencyPair.id)
                                   .join(Exchange, ExchangeCurrencyPair.exchange_id == Exchange.id)
                                   .join(first, ExchangeCurrencyPair.first_id == first.id)
                                   .join(second, ExchangeCurrencyPair.second_id == second.id)
            )
        ),
        metadata=Base.metadata
    )


class HistoricRateView(Base):
    """
    View for Historic-Rates.
    Instead of only showing the ID of the ExchangeCurrencyPair the View displays
    the exchange name and the name of the first and second currency.
    """
    first = aliased(Currency)
    second = aliased(Currency)
    __table__ = create_view(
        name="historic_rates_view",
        selectable=select(
            [
                Exchange.name.label("exchange"),
                first.name.label("first_currency"),
                second.name.label("second_currency"),
                HistoricRate.time,
                HistoricRate.open,
                HistoricRate.high,
                HistoricRate.low,
                HistoricRate.close,
                HistoricRate.volume,
                # HistoricRate.base_volume,
                HistoricRate.market_cap,
            ],
            from_obj=(
                HistoricRate.__table__.join(ExchangeCurrencyPair,
                                            HistoricRate.exchange_pair_id == ExchangeCurrencyPair.id)
                                      .join(Exchange, ExchangeCurrencyPair.exchange_id == Exchange.id)
                                      .join(first, ExchangeCurrencyPair.first_id == first.id)
                                      .join(second, ExchangeCurrencyPair.second_id == second.id)
            )
        ),
        metadata=Base.metadata
    )


DatabaseTable = Union[Exchange, Currency, ExchangeCurrencyPair, Ticker, HistoricRate, Trade, OrderBook]

DatabaseTableType = Union[Type[Exchange], Type[Currency], Type[ExchangeCurrencyPair], Type[Ticker], Type[HistoricRate],
                          Type[Trade], Type[OrderBook]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Contains base validators and reports to represent validation results.

Classes:
    - Validator
    - CompositeValidator
    - ProcessingValidator
    - Report
    - CompositeReport
"""

from __future__ import annotations

import abc
import re
import textwrap
from typing import Text, Union, Any


class Validator(abc.ABC):
    """
    Base class for all validators.

    It offers functionality to validate a given value and to create a report of the validation process.

    Attributes:
        value:
            The value to get validated.
        report:
            The report created during validation.
    """

    def __init__(self, value: Union[Any, Validator]):
        """
        Constructor of Validator.

        @param value: The value to get validated or a validator having the result value.
        """
        if isinstance(value, Validator):
            self.value = value.get_result_value()
        else:
            self.value = value

        self.report = None

    def get_result_value(self) -> Any:
        """
        Gets the value of the validator.

        @return: The value of the Validator.
        """
        return self.value

    @abc.abstractmethod
    def validate(self) -> bool:
        """
        Validates the value and produces a report of the validation process.

        @return: True if the value is valid, False otherwise.
        """
        raise NotImplementedError

    def __bool__(self) -> bool:
        """
        @return: Returns the boolean value of its validation report.
        """
        return bool(self.report)


class CompositeValidator(Validator):
    """
    A composite validator that combines multiple validators.

    Attributes:
        validators:
            A List of children Validators.
    """

    def __init__(self, value: Union[Any, Validator], *child_validators: Validator):
        """
        Constructor of CompositeValidator.

        @param value: The value to get validated.
        @param child_validators: A variable-length sequence of children Validators.
        """
        super().__init__(value)

        self.validators = list(child_validators)
        self.report = CompositeReport()

    def append_validator(self, validator: Validator) -> None:
        """
        Adds a validator to the composite validator.

        @param validator: The validator to be added to the composite validator.
        """
        self.validators.append(validator)

    def append_report(self, report: Union[Validator, Report]) -> None:
        """
        Appends a Report or the Report of a Validator to the CompositeReport.

        @param report: A Validator whose Report will get appended or a Report.
        """
        assert isinstance(self.report, CompositeReport)
        self.report.append_report(report)

    def get_result_value(self) -> Any:
        """
        Returns the value or result.

        Returns its value if no special Validator is the last of its children.
        Otherwise the result value of the last Validator is returned.

        @return: The value or result value.
        """
        last_validator = self.validators[-1]

        if isinstance(last_validator, (ProcessingValidator, CompositeValidator)):
            return last_validator.get_result_value()

        return self.value

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        assert isinstance(self.report, CompositeReport)

        for validator in self.validators:
            is_valid = validator.validate()
            self.report.append_report(validator.report)

            if not is_valid:
                return False

        return True


class ProcessingValidator(Validator):
    """
    Special Validator for processing a value.

    It produces a result value from the given value.

    Attributes:
        result:
            The result produced from the initial value.
    """

    def __init__(self, value: Any):
        """
        Constructor of ProcessingValidator.

        @param value: The value to get validated.
        """
        super().__init__(value)

        self.result = None

    @abc.abstractmethod
    def process(self) -> Any:
        """
        Processes the value.

        Returns the result value from processing the initial value.

        @return: The result value.
        """
        raise NotImplementedError

    def get_result_value(self) -> Any:
        """
        Gets the result value.

        @return: The result value.
        """
        return self.result

    def validate(self) -> bool:
        """
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        self.result = self.process()
        return True


class _Valid:
    """
    Wrapper class for Exception or Text.

    Wrapper class for wrapping Exceptions or Text while providing a
    boolean value (False for Exceptions, True otherwise).

    Attributes:
        message:
            An Exception instance or a Text message (str).
    """

    def __init__(self, message: Union[Exception, Text]):
        """
        Construct a Valid object wrapping the specified message.

        @param message: An exception or text message to be wrapped.
        """
        self.message = message

    def __bool__(self) -> bool:
        """
        @return: False if a exception is wrapped, True otherwise.
        """
        return not isinstance(self.message, Exception)

    def __str__(self) -> Text:
        """
        A method for representing a message in text format.

        A text value returning "-" and the respective message, if message is a
        Exception, otherwise "+" and the respective message.

        @return: "-" for Exceptions, "+" otherwise.
        """
        sign = "+" if self else "-"
        return sign + " " + str(self.message)


class Report:
    """
    Report of Validation.

    Report of the Validation carried out in a Validator.
    Contains a list of Exceptions and Messages collected during Validation.

    Attributes:
        messages:
            List of wrapped Exceptions or Messages.
    """

    def __init__(self, *messages: Union[Exception, Text]):
        """
        Constructor of Report.

        @param messages: A variable-length sequence of Valid objects.
        """
        self.messages = [_Valid(message) for message in messages]

    def indented_report(self) -> Text:
        """
        Indents the generated report representation according to the
        opening and closing brackets.

        @return: The indented representation.
        """
        report = str(self)

        report = re.sub(r"\[", "[\n", report)
        report = re.sub(r",", ",\n", report)
        report = re.sub(r"]", "\n]", report)

        indented_lines = list()
        level = 0

        for line in report.splitlines():
            line = line.strip()

            if "]" in line:
                level -= 1

            indented_line = textwrap.indent(line, level * "  ")
            indented_lines.append(indented_line)

            if "[" in line:
                level += 1

        return "\n".join(indented_lines)

    def print_report(self) -> None:
        """
        Prints the Report's indented representation.
        """
        print(self.indented_report())

    def __bool__(self) -> bool:
        """
        A boolean value returning True if all elements in messages are True,
        otherwise returning False.

        @return: True if all messages are True, False otherwise.
        """
        return all(self.messages)

    def __str__(self) -> Text:
        """
        A method for representing a message in text format.

        A text value returning "-" and the respective message, if message is a
        Exception, otherwise "+" and the respective message.

        @return: "-" for Exceptions, "+" otherwise.
        """
        if len(self.messages) == 1:
            return str(self.messages[0])

        sign = "+" if self else "-"
        return sign + " " + str(self.messages)

    def __repr__(self) -> str:
        """
        A method for representing a message in text format.

        A text value returning "-" and the respective message, if message is a
        Exception, otherwise "+" and the respective message.

        @return: "-" for Exceptions, "+" otherwise.
        """
        return self.__str__()


class CompositeReport(Report):
    """
    CompositeReport consisting of multiple Reports.

    Attributes:
        reports:
            A list of child Reports.
    """

    def __init__(self, *reports: Report):
        """
        Constructor of CompositeReport.

        @param reports: A variable-length sequence of Report objects.
        """
        super().__init__()

        self.reports = list(reports)

    def append_report(self, report: Union[Report, Validator]) -> None:
        """
        Appends a Report or a Validator's Report to children reports.

        @param report: A Report or Validator containing a Report.
        """
        if isinstance(report, Validator):
            report = report.report

        self.reports.append(report)

    def __bool__(self) -> bool:
        """
        A boolean value returning True if all elements in reports are True,
        otherwise returning False.

        @return: True if all reports are True, False otherwise.
        """
        return all(self.reports)

    def __str__(self) -> Text:
        """
        A method for representing a report in text format.

        A text value returning "-" and the respective report, if report is a
        Exception, otherwise "+" and the respective report.

        @return: "-" for Exceptions, "+" otherwise.
        """
        sign = "+" if self else "-"
        return sign + " " + str(self.reports)

    def __len__(self) -> int:
        """
        Gets the report count.

        @return: The report count.
        """
        return len(self.reports)
